# Copyright 2020 by B. Knueven, D. Mildebrath, C. Muir, J-P Watson, and D.L. Woodruff
# This software is distributed under the 3-clause BSD License.
from mpisppy.extensions.extension import Extension
from mpisppy.utils.sputils import find_active_objective
from pyomo.repn.standard_repn import generate_standard_repn
from pyomo.core.expr.numeric_expr import LinearExpression
from mpisppy import global_toc

import pyomo.environ as pyo
import sys
import math
import mpisppy.utils.sputils as sputils
import numpy as np
import mpi4py.MPI as mpi

class CrossScenarioExtension(Extension):
    def __init__(self, spbase_object):
        super().__init__(spbase_object)

        opt = self.opt
        if 'cross_scen_options' in opt.options and \
                'check_bound_improve_iterations' in opt.options['cross_scen_options']:
            self.check_bound_iterations = opt.options['cross_scen_options']['check_bound_improve_iterations']
        else:
            self.check_bound_iterations = None

        self.cur_ib = None
        self.iter_at_cur_ib = 1

        self.cur_ob = None

        self.reenable_W = None
        self.reenable_prox = None

        self.any_cuts = False
        self.iter_since_last_check = 0

    def _disable_W_and_prox(self):
        assert self.reenable_W is None
        assert self.reenable_prox is None
        ## hold the PH object harmless
        opt = self.opt
        self.reenable_W = False
        self.reenable_prox = False
        if not opt.W_disabled and not opt.prox_disabled:
            opt.disable_W_and_prox()
            self.reenable_W = True
            self.reenable_prox = True 
        elif not opt.W_disabled:
            opt._disable_W()
            self.eenable_W = True
        elif not opt.prox_disabled:
            opt._disable_prox()
            self.reenable_prox = True 

    def _enable_W_and_prox(self):
        assert self.reenable_W is not None
        assert self.reenable_prox is not None
        
        opt = self.opt
        if self.reenable_W and self.reenable_prox:
            opt.reenable_W_and_prox()
        elif self.reenable_W:
            opt._reenable_W()
        elif self.reenable_prox:
            opt._reenable_prox()

        self.reenable_W = None
        self.reenable_prox = None

    def _check_bound(self):
        opt = self.opt

        chached_ph_obj = dict()

        for k,s in opt.local_subproblems.items():
            phobj = find_active_objective(s)
            phobj.deactivate()
            chached_ph_obj[k] = phobj
            s._mpisppy_model.EF_Obj.activate()

        teeme = (
            "tee-rank0-solves" in opt.options
             and opt.options["tee-rank0-solves"]
        )
        opt.solve_loop(
                solver_options=opt.current_solver_options,
                dtiming=opt.options["display_timing"],
                gripe=True,
                disable_pyomo_signal_handling=False,
                tee=teeme,
                verbose=opt.options["verbose"],
        )

        local_obs = np.fromiter((s._mpisppy_data.outer_bound for s in opt.local_subproblems.values()),
                                dtype="d", count=len(opt.local_subproblems))

        local_ob = np.empty(1)
        if opt.is_minimizing:
            local_ob[0] = local_obs.max()
        else:
            local_ob[0] = local_obs.min()

        global_ob = np.empty(1)

        if opt.is_minimizing:
            opt.mpicomm.Allreduce(local_ob, global_ob, op=mpi.MAX)
        else: 
            opt.mpicomm.Allreduce(local_ob, global_ob, op=mpi.MIN)

        #print(f"CrossScenarioExtension OB: {global_ob[0]}")

        opt.spcomm.BestOuterBound = opt.spcomm.OuterBoundUpdate(global_ob[0], char='C')

        for k,s in opt.local_subproblems.items():
            s._mpisppy_model.EF_Obj.deactivate()
            chached_ph_obj[k].activate()

    def pre_iter0(self):
        if self.opt.multistage:
            raise RuntimeError("CrossScenarioExtension does not support "
                               "multi-stage problems at this time")
        ## hack as this provides logs for outer bounds
        self.opt.spcomm.has_outerbound_spokes = True

    def post_iter0(self):
        opt = self.opt
        # NOTE: the LShaped code negates the objective, so
        #       we do the same here for consistency
        if 'cross_scen_options' in opt.options and \
                'valid_eta_bound' in opt.options['cross_scen_options']:
            valid_eta_bound = opt.options['cross_scen_options']['valid_eta_bound']
            if not opt.is_minimizing:
                _eta_init = { k: -v for k,v in valid_eta_bound.items() }
            else:
                _eta_init = valid_eta_bound
            _eta_bounds = lambda m,k : (_eta_init[k], None)
        else:
            lb = (-sys.maxsize - 1) * 1. / len(opt.all_scenario_names)
            _eta_init = lambda m,k : lb
            _eta_bounds = lambda m,k : (lb, None)

        # eta is attached to each subproblem, regardless of bundles
        bundling = opt.bundling
        for k,s in opt.local_subproblems.items():
            s._mpisppy_model.eta = pyo.Var(opt.all_scenario_names, initialize=_eta_init, bounds=_eta_bounds)
            if sputils.is_persistent(s._solver_plugin):
                for var in s._mpisppy_model.eta.values():
                    s._solver_plugin.add_var(var)
            if bundling: ## create a refence to eta on each subproblem
                for sn in s.scen_list:
                    scenario = opt.local_scenarios[sn]
                    scenario._mpisppy_model.eta = { k : s._mpisppy_model.eta[k] for k in opt.all_scenario_names }

        ## hold the PH object harmless
        self._disable_W_and_prox()
        
        for k,s in opt.local_subproblems.items():

            obj = find_active_objective(s)

            repn = generate_standard_repn(obj.expr, quadratic=True)
            if len(repn.nonlinear_vars) > 0:
                raise ValueError("CrossScenario does not support models with nonlinear objective functions")

            if bundling:
                ## NOTE: this is slighly wasteful, in that for a bundle
                ##       the first-stage cost appears len(s.scen_list) times
                ##       If this really made a difference, we could use s.ref_vars
                ##       to do the substitution
                nonant_vardata_list = list()
                for sn in s.scen_list:
                    nonant_vardata_list.extend( \
                            opt.local_scenarios[sn]._mpisppy_node_list[0].nonant_vardata_list)
            else:
                nonant_vardata_list = s._mpisppy_node_list[0].nonant_vardata_list

            nonant_ids = set((id(var) for var in nonant_vardata_list))

            linear_coefs = list(repn.linear_coefs)
            linear_vars = list(repn.linear_vars)

            quadratic_coefs = list(repn.quadratic_coefs)

            # adjust coefficients by scenario/bundle probability
            scen_prob = s._mpisppy_probability
            for i,var in enumerate(repn.linear_vars):
                if id(var) not in nonant_ids:
                    linear_coefs[i] *= scen_prob

            for i,(x,y) in enumerate(repn.quadratic_vars):
                # only multiply through once
                if id(x) not in nonant_ids:
                    quadratic_coefs[i] *= scen_prob
                elif id(y) not in nonant_ids:
                    quadratic_coefs[i] *= scen_prob

            # NOTE: the LShaped code negates the objective, so
            #       we do the same here for consistency
            if not opt.is_minimizing:
                for i,coef in enumerate(linear_coefs):
                    linear_coefs[i] = -coef
                for i,coef in enumerate(quadratic_coefs):
                    quadratic_coefs[i] = -coef

            # add the other etas
            if bundling:
                these_scenarios = set(s.scen_list)
            else:
                these_scenarios = [k]

            eta_scenarios = list()
            for sn in opt.all_scenario_names:
                if sn not in these_scenarios:
                    linear_coefs.append(1)
                    linear_vars.append(s._mpisppy_model.eta[sn])
                    eta_scenarios.append(sn)

            expr = LinearExpression(constant=repn.constant, linear_coefs=linear_coefs,
                                    linear_vars=linear_vars)

            if repn.quadratic_vars:
                expr += pyo.quicksum(
                    (coef*x*y for coef,(x,y) in zip(quadratic_coefs, repn.quadratic_vars))
                )

            s._mpisppy_model.EF_obj = pyo.Expression(expr=expr)

            if opt.is_minimizing:
                s._mpisppy_model.EF_Obj = pyo.Objective(expr=s._mpisppy_model.EF_obj, sense=pyo.minimize)
            else:
                s._mpisppy_model.EF_Obj = pyo.Objective(expr=-s._mpisppy_model.EF_obj, sense=pyo.maximize)
            s._mpisppy_model.EF_Obj.deactivate()

            # add cut constraint dicts
            s._mpisppy_model.benders_cuts = pyo.Constraint(pyo.Any)
            s._mpisppy_model.inner_bound_constr = pyo.Constraint(pyo.Any)

        self._enable_W_and_prox()

        # try to get the initial eta LB cuts
        # (may not be available)
        opt.spcomm.get_from_cross_cuts()

    def miditer(self):
        self.iter_since_last_check += 1

        ib = self.opt.spcomm.BestInnerBound
        if ib != self.cur_ib:
            self.cur_ib = ib
            self.iter_at_cur_ib = 1
        elif self.cur_ib is not None and math.isfinite(self.cur_ib):
            self.iter_at_cur_ib += 1

        ob = self.opt.spcomm.BestOuterBound
        if self.cur_ob is not None and math.isclose(ob, self.cur_ob):
            ob_new = False
        else:
            self.cur_ob = ob
            ob_new = True
        
        if not self.any_cuts:
            if self.opt.spcomm.new_cuts:
                self.any_cuts = True

        ## if its the second time or more with this IB, we'll only check
        ## if the last improved the OB, or if the OB is new itself (from somewhere else)
        check = (self.check_bound_iterations is not None) and self.any_cuts and ( \
                (self.iter_at_cur_ib == self.check_bound_iterations) or \
                (self.iter_at_cur_ib > self.check_bound_iterations and ob_new) or \
                ((self.iter_since_last_check%self.check_bound_iterations == 0) and self.opt.spcomm.new_cuts))
                # if there hasn't been OB movement, check every so often if we have new cuts
        if check:
            global_toc(f"Attempting to update Best Bound with CrossScenarioExtension")
            self._check_bound()
            self.opt.spcomm.new_cuts = False
            self.iter_since_last_check = 0

    def enditer(self):
        pass

    def post_everything(self):
        pass



# Copyright 2020 by B. Knueven, D. Mildebrath, C. Muir, J-P Watson, and D.L. Woodruff
# This software is distributed under the 3-clause BSD License.
"""
This is the logging configuration for mpisppy.

The documentation below is primarily for mpisppy developers.

Examples
========
To use the logger in your code, add the following 
after your import
.. code-block:: python
   
   import logging
   logger = logging.getLogger('mpisppy.path.to.module')

Then, you can use the standard logging functions
.. code-block:: python
   
   logger.debug('message')
   logger.info('message')
   logger.warning('message')
   logger.error('message')
   logger.critical('message')
   
Note that by default, any message that has a logging level
of warning or higher (warning, error, critical) will be
logged.

To log an exception and capture the stack trace
.. code-block:: python

   try:
      c = a / b
   except Exception as e:
      logging.error("Exception occurred", exc_info=True)

"""
import sys
import logging
log_format = '%(message)s'

# configure the root logger for mpisppy
logger = logging.getLogger('mpisppy')
logger.setLevel(logging.INFO)

console_handler = logging.StreamHandler(sys.stdout)
fmtr = logging.Formatter(log_format)
console_handler.setFormatter(fmtr)
logger.addHandler(console_handler)

def setup_logger(name, out, level=logging.DEBUG, mode='w', fmt=None):
    ''' Set up a custom logger quickly
        https://stackoverflow.com/a/17037016/8516804
    '''
    if fmt is None:
        fmt = "(%(asctime)s) %(message)s"
    l = logging.getLogger(name)
    l.setLevel(level)
    l.propagate = False
    formatter = logging.Formatter(fmt)
    if out in (sys.stdout, sys.stderr):
        handler = logging.StreamHandler(out)
    else: # out is a filename
        handler = logging.FileHandler(out, mode=mode) 
    handler.setFormatter(formatter)
    l.addHandler(handler)



# Amalgomator.py starting point; DLW March 2021
# Copyright 2021 by D.L. Woodruff
# This software is distributed under the 3-clause BSD License.
# To test : python amalgomator.py --num-scens=10

"""Takes a scenario list and a scenario creator (and options)
as input and produces an outer bound on the objective function (solving the EF directly
or by decomposition). This same object should also be able to return an x-hat. 

This thing basically wraps the functionality of the "standard" *_cylinder examples.

It may be an extenisble base class, but not abstract.


Wait. I guess I need a function that takes an argparse object and returns the kwargs dict for the
scenario creator function and I need to be passed in a starting argparse object that
has the needed stuff for the problem at hand.

"""

"""
Input options:
EF|PH|APH|L, etc
Assuming not EF:
  - list of cylinders
  - list of hub extensions

The object will then call the appropriate baseparsers functions to set up
the args and assemble the objects needed for spin-the-wheel, which it will call.

[NOTE: argparse can be passed a command list and can generate a command list that can be passed]


Things that will be needed to support this:
  - mapping from cylinder name to class?
  - mapping from extension name to class?
  = The same kind of mapping to the baseparser functions
  = Support in baseparsers for "all" cylinders and extensions with a mapping to the function
 ==== we can't keep hanging at the termination barrier!

*mapping*: I think the cylinder "name" and extension "name" should just be the module
name to import in string from (e.g., "mpisppy.opt.ph") and then use importlib to 
import it.

? Should baseparser functions "register themselves" in a import-name-string: function
map that is global to baseparsers (i.e. importable)?


"""
import numpy as np
import importlib
import pyomo.environ as pyo
import argparse

from mpisppy.utils.sputils import spin_the_wheel, get_objs, nonant_cache_from_ef
import mpisppy.utils.baseparsers as baseparsers
import mpisppy.utils.sputils as sputils
from mpisppy.utils import vanilla
from mpisppy import global_toc

#==========
def from_module(mname, options, extraargs=None, use_command_line=True):
    """ Try to get everything from one file (this will not always be possible).
    Args:
        mname (str): the module name (module must have certain functions)
        options (dict): Amalgomator options or extra arguments to use 
                        in addition with the commande line
        extraargs (ArgumentParser) : extra arguments to specify in the command line, e.g. for MMW
        use_command_line (bool): should we take into account the command line to add options ?
                                 default is True
                    
    Returns:
        ama (Amalgomator): the instantiated object
    
    Note :
        Use adict iif you want to bypass the command line.
    """
    everything = ["scenario_names_creator",
                 "scenario_creator",
                 "inparser_adder",
                 "kw_creator"]  # start and denouement can be missing.

    assert(options is not None)
    m = importlib.import_module(mname)

    you_can_have_it_all = True
    for ething in everything:
        if not hasattr(m, ething):
            print(f"Module {mname} is missing {ething}")
            you_can_have_it_all = False
    if not you_can_have_it_all:
        raise RuntimeError(f"Module {mname} not complete for from_module")
        
    options = Amalgomator_parser(options, m.inparser_adder,
                                 extraargs=extraargs,
                                 use_command_line=use_command_line)
    options['_mpisppy_probability'] = 1/options['num_scens']
    start = options['start'] if(('start' in options)) else None
    sn = m.scenario_names_creator(options['num_scens'], start=start)
    dn = m.scenario_denouement if hasattr(m, "scenario_denouement") else None
    ama = Amalgomator(options,
                      sn,
                      m.scenario_creator,
                      m.kw_creator,
                      scenario_denouement=dn)
    return ama


#==========
def _bool_option(options, oname):
    return oname in options and options[oname]


#==========
def Amalgomator_parser(options, inparser_adder, extraargs=None, use_command_line=True):
    """ Helper function for Amalgomator.  This gives us flexibility (e.g., get scen count)
    Args:
        options (dict): Amalgomator control options
        inparser_adder (fct): returns updated ArgumentParser the problem
        extraargs (ArgumentParser) : extra arguments to specify in the command line, e.g. for MMW
        use_command_line (bool): should we take into account the command line to add options ?
                                 default is True
    Returns;
        options (dict): a dict containing the options, both parsed values and pre-set options
    """

    opt = {**options}
    
    if use_command_line:
        if _bool_option(options, "EF-2stage"):
            parser = baseparsers.make_EF2_parser(num_scens_reqd=_bool_option(options, "num_scens_reqd"))
        else:
            raise RuntimeError("only EF-2Stage is supported right now")
    
        # TBD add args for everything else that is not EF, which is a lot
    
        # call inparser last (so it can delete args if it needs to)
        inparser_adder(parser)
        
        if extraargs is not None:
            parser = argparse.ArgumentParser(parents = [parser,extraargs],conflict_handler='resolve')
        
        args = parser.parse_args()
    
        opt.update(vars(args)) #Changes made via the command line overwrite what is in options 
        
        if not ('EF_solver_name' in opt):
            opt["EF_solver_options"]["mipgap"] = opt["EF_mipgap"]
        else:
            opt["EF_solver_options"] = {"mipgap": opt["EF_mipgap"]}
    
    else:
        #Checking if options has all the options we need 
        if not ('EF_solver_name' in opt):
            opt['EF_solver_name'] = "gurobi"
        if not ('EF_solver_options' in opt):
            opt['EF_solver_options'] = {'mipgap': None}
        if not ('num_scens' in opt):
            raise RuntimeWarning("options should have a number of scenarios to compute a xhat")
        
      
    return opt
    

#========================================
class Amalgomator():
    """Takes a scenario list and a scenario creator (and options)
    as input. The ides is to produce an outer bound on the objective function (solving the EF directly
    or by decomposition) and/or an x-hat with inner bound; however, what it does is controlled by
    its constructor options and by user options.

    This thing basically wraps the functionality of the "standard" *_cylinder examples.
    
    It may be an extenisble base class, but not abstract.

    Args:
        options (dict): controls the amalgomation
        scenario_names (list of str) the full set of scenario names
        scenario_creator (fct): returns a concrete model with special things
        kw_creator (fct): takes an args object and returns scenario_creator kwargs
        scenario_denouement (fct): (optional) called at conclusion
    """

    def __init__(self, options,
                 scenario_names, scenario_creator, kw_creator, 
                 scenario_denouement=None, verbose=True):
        self.options = options
        self.scenario_names = scenario_names
        self.scenario_creator = scenario_creator
        self.scenario_denouement = scenario_denouement
        self.kw_creator = kw_creator
        self.verbose = verbose
        self.is_EF = _bool_option(options, "EF-2stage") or _bool_option(options, "EF-mstage")
        if self.is_EF:
            self.solvername = options['EF_solver_name'] if  ('EF_solver_name' in options) else 'gurobi'
            self.solver_options = options['EF_solver_options'] \
                if ('EF_solver_options' in options) else {}
        
    def run(self):
        
        """ Top-level execution."""
        if self.is_EF:
            kwargs = self.kw_creator(self.options)     
            
            ef = sputils.create_EF(
                self.scenario_names,
                self.scenario_creator,
                scenario_creator_kwargs=kwargs,
                suppress_warnings=True,
            )
            
            solvername = self.solvername
            solver = pyo.SolverFactory(solvername)
            if hasattr(self, "solver_options") and (self.solver_options is not None):
                for option_key,option_value in self.solver_options.items():
                    if option_value is not None:
                        solver.options[option_key] = option_value
            if self.verbose :
                global_toc("Starting EF solve")
            if 'persistent' in solvername:
                solver.set_instance(ef, symbolic_solver_labels=True)
                results = solver.solve(tee=False)
            else:
                results = solver.solve(ef, tee=False, symbolic_solver_labels=True,)
            if self.verbose:
                global_toc("Completed EF solve")

            
            self.EF_Obj = pyo.value(ef.EF_Obj)

            objs = get_objs(ef)
            
            self.is_minimizing = objs[0].is_minimizing
            #TBD : Write a function doing this
            if self.is_minimizing:
                self.best_outer_bound = results.Problem[0]['Lower bound']
                self.best_inner_bound = results.Problem[0]['Upper bound']
            else:
                self.best_inner_bound = results.Problem[0]['Upper bound']
                self.best_outer_bound = results.Problem[0]['Lower bound']
            self.ef = ef

        else:
            self.ef = None   # ???? do we want to retain these objects?
            raise RuntimeError("We can only do EF right now")


if __name__ == "__main__":
    # for debugging
    import mpisppy.tests.examples.farmer as farmer
    # EF, PH, L-shaped, APH flags, and then boolean multi-stage
    ama_options = {"EF-2stage": True}   # 2stage vs. mstage
    
    ama = from_module("mpisppy.tests.examples.farmer", ama_options)
    ama.run()
    print(f"inner bound=", ama.best_inner_bound)
    print(f"outer bound=", ama.best_outer_bound)
    print(nonant_cache_from_ef(ama.ef))
    # wish list: allow for integer relaxation using the Pyomo inplace transformation
    """ Issues:
    0. What do we want to put in ama_options and what do we want to discover
       by looking at data? I am inclined to put things in the data and maybe
       check against data. Here are the things:
       - 2-stage versus multi-stage
       - MIP versus continuous only
       - linear/quadratic versus non-linear
    """



# Copyright 2020 by B. Knueven, D. Mildebrath, C. Muir, J-P Watson, and D.L. Woodruff
# This software is distributed under the 3-clause BSD License.
from pyomo.core.base.block import _BlockData, declare_custom_block
import pyomo.environ as pe
from pyomo.solvers.plugins.solvers.persistent_solver import PersistentSolver
import pyomo.contrib.benders.benders_cuts as bc

try:
    from mpi4py import MPI

    mpi4py_available = True
except:
    mpi4py_available = False
try:
    import numpy as np

    numpy_available = True
except:
    numpy_available = False
import logging

logger = logging.getLogger(__name__)


solver_dual_sign_convention = dict()
solver_dual_sign_convention['ipopt'] = -1
solver_dual_sign_convention['gurobi'] = -1
solver_dual_sign_convention['gurobi_direct'] = -1
solver_dual_sign_convention['gurobi_persistent'] = -1
solver_dual_sign_convention['cplex'] = -1
solver_dual_sign_convention['cplex_direct'] = -1
solver_dual_sign_convention['cplexdirect'] = -1
solver_dual_sign_convention['cplex_persistent'] = -1
solver_dual_sign_convention['glpk'] = -1
solver_dual_sign_convention['cbc'] = -1
solver_dual_sign_convention['xpress_direct'] = -1
solver_dual_sign_convention['xpress_persistent'] = -1


@declare_custom_block(name='LShapedCutGenerator')
class LShapedCutGeneratorData(bc.BendersCutGeneratorData):
    def __init__(self, component):
        super().__init__(component)
        # self.local_subproblem_count = 0
        # self.global_subproblem_count = 0

    def set_ls(self, ls):
        self.ls = ls
        self.global_subproblem_count = len(self.ls.all_scenario_names)
        self._subproblem_ndx_map = dict.fromkeys(range(len(self.ls.local_scenario_names)))
        for s in self._subproblem_ndx_map.keys():
            self._subproblem_ndx_map[s] = self.ls.all_scenario_names.index(self.ls.local_scenario_names[s])
        # print(self._subproblem_ndx_map)
        self.all_root_etas = list(self.ls.root.eta.values())

    def global_num_subproblems(self):
        return self.global_subproblem_count

    def add_subproblem(self, subproblem_fn, subproblem_fn_kwargs, root_eta, subproblem_solver='gurobi_persistent',
                       relax_subproblem_cons=False, subproblem_solver_options=None):
        # print(self._subproblem_ndx_map)
        # self.all_root_etas.append(root_eta)
        # self.global_subproblem_count += 1
        if subproblem_fn_kwargs['scenario_name'] in self.ls.local_scenario_names:
            # self.local_subproblem_count += 1
            self.root_etas.append(root_eta)
            subproblem, complicating_vars_map = subproblem_fn(**subproblem_fn_kwargs)
            self.subproblems.append(subproblem)
            self.complicating_vars_maps.append(complicating_vars_map)
            bc._setup_subproblem(subproblem, root_vars=[complicating_vars_map[i] for i in self.root_vars if
                                                       i in complicating_vars_map],
                              relax_subproblem_cons=relax_subproblem_cons)

            # self._subproblem_ndx_map[self.local_subproblem_count - 1] = self.global_subproblem_count - 1

            if isinstance(subproblem_solver, str):
                subproblem_solver = pe.SolverFactory(subproblem_solver)
            self.subproblem_solvers.append(subproblem_solver)
            if isinstance(subproblem_solver, PersistentSolver):
                subproblem_solver.set_instance(subproblem)
            if subproblem_solver_options:
                for k,v in subproblem_solver_options.items():
                    subproblem_solver.options[k] = v



# Copyright 2020 by B. Knueven, D. Mildebrath, C. Muir, J-P Watson, and D.L. Woodruff
# This software is distributed under the 3-clause BSD License.
import os
import time
import logging
import random
import mpisppy.log
import mpisppy.utils.sputils as sputils
import mpisppy.cylinders.spoke as spoke
import mpi4py.MPI as mpi

from math import inf, isclose
from mpisppy.utils.xhat_eval import Xhat_Eval
from mpisppy.extensions.xhatbase import XhatBase

# Could also pass, e.g., sys.stdout instead of a filename
mpisppy.log.setup_logger("mpisppy.cylinders.xhatshufflelooper_bounder",
                         "xhatclp.log",
                         level=logging.CRITICAL)                         
logger = logging.getLogger("mpisppy.cylinders.xhatshufflelooper_bounder")

class XhatShuffleInnerBound(spoke.InnerBoundNonantSpoke):

    converger_spoke_char = 'X'

    def xhatbase_prep(self):
        if self.opt.multistage:
            raise RuntimeError('The XhatShuffleInnerBound only supports '
                               'two-stage models at this time.')

        verbose = self.opt.options['verbose']
        if "bundles_per_rank" in self.opt.options\
           and self.opt.options["bundles_per_rank"] != 0:
            raise RuntimeError("xhat spokes cannot have bundles (yet)")

        # Start code to support running trace. TBD: factor this up?
        if self.cylinder_rank == 0 and \
                'suffle_running_trace_prefix' in self.opt.options and \
                self.opt.options['shuffle_running_trace_prefix'] is not None:
            running_trace_prefix =\
                            self.opt.options['shuffle_running_trace_prefix']

            filen = running_trace_prefix+self.__class__.__name__+'.csv'
            if os.path.exists(filen):
                raise RuntimeError(f"running trace file {filen} already exists!")
            with open(filen, 'w') as f:
                f.write("time,scen,value\n")
            self.running_trace_filen = filen
        else:
            self.running_trace_filen = None
        # end code to support running trace
        
        if not isinstance(self.opt, Xhat_Eval):
            raise RuntimeError("XhatShuffleInnerBound must be used with Xhat_Eval.")
            
        xhatter = XhatBase(self.opt)

        ### begin iter0 stuff
        xhatter.pre_iter0()
        self.opt._save_original_nonants()

        teeme = False
        if "tee-rank0-solves" in self.opt.options:
            teeme = self.opt.options['tee-rank0-solves']

        self.opt.solve_loop(
            solver_options=self.opt.current_solver_options,
            dtiming=False,
            gripe=True,
            tee=teeme,
            verbose=verbose
        )
        self.opt._update_E1()  # Apologies for doing this after the solves...
        if abs(1 - self.opt.E1) > self.opt.E1_tolerance:
            raise ValueError(f"Total probability of scenarios was {self.opt.E1}"+\
                                 f"E1_tolerance = {self.opt.E1_tolerance}")
        infeasP = self.opt.infeas_prob()
        if infeasP != 0.:
            raise ValueError(f"Infeasibility detected; E_infeas={infeasP}")

        ### end iter0 stuff

        xhatter.post_iter0()
        self.opt._save_nonants() # make the cache

        ## for later
        self.verbose = self.opt.options["verbose"] # typing aid  
        self.solver_options = self.opt.options["xhat_looper_options"]["xhat_solver_options"]
        self.xhatter = xhatter

        ## option drive this? (could be dangerous)
        self.random_seed = 42
        # Have a separate stream for shuffling
        self.random_stream = random.Random()


    def try_scenario(self, scenario):
        obj = self.xhatter._try_one({"ROOT":scenario},
                            solver_options = self.solver_options,
                            verbose=False,
                            restore_nonants=False)
        def _vb(msg): 
            if self.verbose and self.opt.cylinder_rank == 0:
                print ("(rank0) " + msg)

        if self.running_trace_filen is not None:
            with open(self.running_trace_filen, "a") as f:
                f.write(f"{time.time()},{scenario},{obj}\n")
        if obj is None:
            _vb(f"    Infeasible {scenario}")
            return False
        _vb(f"    Feasible {scenario}, obj: {obj}")

        update = self.update_if_improving(obj)
        logger.debug(f'   bottom of try_scenario on rank {self.global_rank}')
        return update

    def main(self):
        verbose = self.opt.options["verbose"] # typing aid  
        logger.debug(f"Entering main on xhatshuffle spoke rank {self.global_rank}")

        self.xhatbase_prep()

        # give all ranks the same seed
        self.random_stream.seed(self.random_seed)
        # shuffle the scenarios (i.e., sample without replacement)
        shuffled_scenarios = self.random_stream.sample(
            self.opt.all_scenario_names,
            len(self.opt.all_scenario_names))
        scenario_cycler = ScenarioCycler(shuffled_scenarios)

        def _vb(msg): 
            if self.verbose and self.opt.cylinder_rank == 0:
                print("(rank0) " + msg)

        xh_iter = 1
        while not self.got_kill_signal():

            if (xh_iter-1) % 100 == 0:
                logger.debug(f'   Xhatshuffle loop iter={xh_iter} on rank {self.global_rank}')
                logger.debug(f'   Xhatshuffle got from opt on rank {self.global_rank}')

            if self.new_nonants:
                # similar to above, not all ranks will agree on
                # when there are new_nonants (in the same loop)
                logger.debug(f'   *Xhatshuffle loop iter={xh_iter}')
                logger.debug(f'   *got a new one! on rank {self.global_rank}')
                logger.debug(f'   *localnonants={str(self.localnonants)}')

                # update the caches
                self.opt._put_nonant_cache(self.localnonants)
                self.opt._restore_nonants()

            next_scenario = scenario_cycler.get_next()
            if next_scenario is not None:
                _vb(f"   Trying next {next_scenario}")
                update = self.try_scenario(next_scenario)
                if update:
                    _vb(f"   Updating best to {next_scenario}")
                    scenario_cycler.best = next_scenario
            else:
                scenario_cycler.begin_epoch()

            #_vb(f"    scenario_cycler._scenarios_this_epoch {scenario_cycler._scenarios_this_epoch}")

            xh_iter += 1


class ScenarioCycler:

    def __init__(self, all_scenario_list):
        self._all_scenario_list = all_scenario_list
        self._num_scenarios = len(all_scenario_list)
        self._cycle_idx = 0
        self._cur_scen = all_scenario_list[0]
        self._scenarios_this_epoch = set()
        self._best = None

    @property
    def best(self):
        self._scenarios_this_epoch.add(self._best)
        return self._best

    @best.setter
    def best(self, value):
        self._best = value

    def begin_epoch(self):
        self._scenarios_this_epoch = set()

    def get_next(self):
        next_scen = self._cur_scen
        if next_scen in self._scenarios_this_epoch:
            self._iter_scen()
            return None
        self._scenarios_this_epoch.add(next_scen)
        self._iter_scen()
        return next_scen

    def _iter_scen(self):
        self._cycle_idx += 1
        ## wrap around
        self._cycle_idx %= self._num_scenarios
        self._cur_scen = self._all_scenario_list[self._cycle_idx]



# Copyright 2020 by B. Knueven, D. Mildebrath, C. Muir, J-P Watson, and D.L. Woodruff
# This software is distributed under the 3-clause BSD License.
import os.path
from inspect import signature
import pyomo.environ as pyo

import mpisppy

from pysp.scenariotree.tree_structure_model import \
        CreateAbstractScenarioTreeModel, ScenarioTreeModelFromNetworkX
from pysp.phutils import isVariableNameIndexed,\
                                extractVariableNameAndIndex
from pyomo.common.fileutils import import_file

from mpisppy.scenario_tree import ScenarioNode as mpisppyScenarioNode
from mpisppy.utils.sputils import create_EF as create_mpisppy_EF

hasnetworkx = False
try:
    import networkx
    hasnetworkx = True
except:
    pass

def _extract_name_and_index(name):
    '''Determine if a variable name is indexed, strip off the
       index, and return the name (without index) and the index.
       The index will be None if no index is found'''
    index = None
    if isVariableNameIndexed(name):
        name, index = extractVariableNameAndIndex(name)
    return name, index

class PySPModel:
    """A class for instantiating PySP models for use in mpisppy. 

    Args:
      scenario_creator (str or fct): 
                    is a path to the file that contains the scenario 
                    callback for concrete or the reference model for abstract,
                    or the scenario creator callback function
      tree_model (concrete model, or networkx tree, or str, or None): 
                    gives the tree as a concrete model 
                    (which could be a fct) or a valid networkx scenario tree
                    or path to AMPL data file.
      scenarios_dir (str or None):
                    For abstract models, gives the directory of the scenarios.
                    If None for an abstract model, will be inferred from the
                    directory of the tree_model
      scenario_creator_callback_name (str or None):
                    The name of the scenario creator callback function
                    if specifying an concrete model module in
                    scenario_creator
      tree_model_callback_name (str or None):
                    The name of the tree model creator callback function
                    if specifying an concrete model module in
                    scenario_creator

    Properties:
      all_scenario_names (list):
                    A list of scenario names base on the pysp model for
                    use in mpisppy
      scenario_creator (fct):
                    A scenario creator function based on the pysp model
                    for use in mpisppy
      scenario_denouement (fct):
                    A blank scenario_denouement function for use in mpisppy
    """
    def __init__(self, scenario_creator, tree_model=None,
            scenarios_dir=None,
            scenario_creator_callback_name=None,
            tree_model_callback_name=None):

        ## first, attempt to determine abstract vs concrete
        ## and get a scenario instance creator

        ## if callable, a instance creator
        if callable(scenario_creator):
            self.pysp_instance_creator = scenario_creator
            self.abstract = False
        else: ## else, either and abstract model or a module with a callback
            if scenario_creator_callback_name is None:
                scenario_creator_callback_name = 'pysp_instance_creation_callback'
            module = import_file(scenario_creator)
            if hasattr(module, scenario_creator_callback_name):
                self.pysp_instance_creator = \
                        getattr(module, scenario_creator_callback_name)
                self.abstract = False
            else:
                self.pysp_instance_creator = module.model.create_instance
                self.abstract = True

        ## attempt to find and construct a tree model
        if tree_model is None:
            if tree_model_callback_name is None:
                tree_model_callback_name = 'pysp_scenario_tree_model_callback'
            tree_maker = getattr(module, tree_model_callback_name)

            tree_model = tree_maker()
        ## if we get a *.dat file, assume the scenarios are here unless
        ## otherwise specified
        if isinstance(tree_model, str):
            self.tree_model = CreateAbstractScenarioTreeModel(\
                                ).create_instance(tree_model)
            self.scenarios_dir = os.path.dirname(tree_model)
        elif hasnetworkx and isinstance(tree_model, networkx.DiGraph):
            self.tree_model = ScenarioTreeModelFromNetworkX(tree_model)
        elif isinstance(tree_model, pyo.ConcreteModel):
            self.tree_model = tree_model
        else:
            raise RuntimeError("Type of tree_model {} unrecongized".format(
                                type(tree_model)))

        ## set the scenarios_dir if specified, but complain if 
        ## we don't have an abstract model
        if scenarios_dir is not None:
            if not self.abstract:
                raise RuntimeError("An abstract model is required for "
                        "scenarios_dir")
            self.scenarios_dir = scenarios_dir

        self._init()
    
    def _init(self):
        """Sets up things needed for mpisppy"""

        ## get the things out of the tree model we need
        tree_model = self.tree_model
        self._scenario_names = list(tree_model.Scenarios)
        
        ## check for more than two stages
        if len(tree_model.Stages) != 2:
            raise RuntimeError("Models for PySPModel must be 2-stage, "\
                    "found {} stages".format(len(tree_model.Stages)))
        ## gripe if we see bundles
        if len(tree_model.Bundles) != 0:
            logger.warning("Bundles are ignored in PySPModel")

        ## extract first stage information from the scenario tree
        first_stage = tree_model.Stages[1]
        first_stage_cost_str = tree_model.StageCost[first_stage].value 

        first_stage_cost_name, first_stage_cost_index = \
                _extract_name_and_index(first_stage_cost_str)

        ## this is the function that becomes the scenario creator callback
        def scenario_creator_callback(scenario_name, **kwargs):
            ## fist, get the model out
            if self.abstract:
                model = self.pysp_instance_creator(
                        os.path.join(self.scenarios_dir,scenario_name+'.dat'))
            else:
                ## try to support both callback types, but pass in Nones for 
                ## args that aren't scenario_name
                # TBD (use inspect to match kwargs with signature)
                try:
                    model = self.pysp_instance_creator(None,
                                                       scenario_name,
                                                       None,
                                                       **kwargs)
                except TypeError:
                    try:
                        model = self.pysp_instance_creator(scenario_name,
                                                           **kwargs)
                    except TypeError:
                        try:
                            model = self.pysp_instance_creator(scenario_name, None)
                            for key,val in kwargs.items():
                                if val is not None:
                                    print("WARNING: did not use {}={}".\
                                          format(key, val))
                        except:
                            print("signature=",
                                  str(signature(self.pysp_instance_creator)))
                            raise RuntimeError("Could not match callback")

            ## extract the first stage cost expression
            stage_cost_expr = getattr(model, first_stage_cost_name)
            if first_stage_cost_index is None:
                cost_expression = stage_cost_expr
            else:
                cost_expression = stage_cost_expr[first_stage_cost_index]

            ## now collect the nonant vars from the model based
            ## on the tree_model
            nonant_list = list()
            for var_str in tree_model.StageVariables[first_stage]:
                var_name, var_index = _extract_name_and_index(var_str)
                pyovar = getattr(model, var_name)
                ## if there's no index, we can append the var
                if var_index is None:
                    nonant_list.append(pyovar)
                ## if there is an index, it should be a single 
                ## index or the whole thing.
                ## NOTE: it would not be too difficult to enable
                ##        slicing, if necessary
                elif isinstance(var_index,tuple):
                    if '*' in var_index:
                        for i in var_index:
                            if i != '*':
                                raise RuntimeError("PySPModel does not "
                                                    "support slicing")
                        nonant_list.append(pyovar)
                ## If we only have one index
                elif var_index == '*':
                    nonant_list.append(pyovar)
                else: ## the index is not a slice
                    nonant_list.append(pyovar[var_index])

            ## get the probability from the tree_model
            scen_prob = tree_model.ConditionalProbability[
                            tree_model.ScenarioLeafNode[scenario_name].value
                            ].value

            ## add the things mpisppy expects to the model
            model._mpisppy_probability = scen_prob

            model._mpisppy_node_list = [mpisppyScenarioNode(
                                        name="ROOT",
                                        cond_prob=1.0,
                                        stage=1,
                                        cost_expression=cost_expression,
                                        scen_name_list=self._scenario_names,
                                        nonant_list = nonant_list,
                                        scen_model=model)
                                    ]
            return model

        ## add this function to the class
        self._mpisppy_instance_creator = scenario_creator_callback

    @property
    def scenario_creator(self):
        return self._mpisppy_instance_creator

    @property
    def all_scenario_names(self):
        return self._scenario_names

    @property
    def scenario_denouement(self):
        return lambda *args,**kwargs: None



# Copyright 2020 by B. Knueven, D. Mildebrath, C. Muir, J-P Watson, and D.L. Woodruff
# This software is distributed under the 3-clause BSD License.
""" fraction of ints converger to illustrate a custom converger that does not
    directly use mpi reduce calls.
    DLW Jan 2019
"""

import math
import pyomo.environ as pyo
import mpisppy.convergers.converger

class FractionalConverger(mpisppy.convergers.converger.Converger):
    """ Illustrate a class to contain data used by the converger 
    NOTE: unlike the extensions, we get only our object back so
          we need to keep references to everything we might want
          to look at.
    Args:
        options (dict): keys are option names
        local_scenarios (dict): keys are names, 
                                vals are concrete models with attachments
        comms (dict): key is node name; val is a comm object
        rank (int): mpi process rank
    """
    def __init__(self, phb):
        options = phb.options
        self.name = "fractintsnotconv"
        self.verbose = options["verbose"]
        self._options = options
        self._local_scenarios = phb.local_scenarios
        self._comms = phb.comms
        self._rank = phb.cylinder_rank
        if self.verbose:
            print ("Created converger=",self.name)
        
    def _convergence_value(self):
        """ compute the fraction of *not* converged ints
        Args:
            self (object): create by prep

        Returns:
            number of converged ints divided by number of ints
        """
        numints = 0
        numconv = 0
        for k,s in self._local_scenarios.items():
            nlens = s._mpisppy_data.nlens        
            for node in s._mpisppy_node_list:
                ndn = node.name
                for i in range(nlens[ndn]):
                    xvar = node.nonant_vardata_list[i]
                    if xvar.is_integer() or xvar.is_binary():
                        numints += 1
                        xb = pyo.value(s._mpisppy_model.xbars[(ndn,i)])
                        #print ("dlw debug",xb*xb, pyo.value(s._mpisppy_model.xsqbars[(ndn,i)]))
                        if math.isclose(xb * xb, pyo.value(s._mpisppy_model.xsqbars[(ndn,i)]), abs_tol=1e-09):
                            numconv += 1
        if self.verbose:
            print (self.name,": numints=",numints)
        if numints > 0:
            retval = 1.0 - numconv / numints
        else:
            retval = 0
        if self.verbose:
            print (self.name,": convergence value=",retval)
        return retval

    def is_converged(self):
        """ check for convergence
        Args:
            self (object): create by prep

        Returns:
           converged?: True if converged, False otherwise
        """
        return self._convergence_value() < self._options['convthresh']



# Copyright 2020 by B. Knueven, D. Mildebrath, C. Muir, J-P Watson, and D.L. Woodruff
# This software is distributed under the 3-clause BSD License.
import unittest
import pyomo.environ as pyo
import os.path

from mpisppy.utils.pysp_model import PySPModel
from mpisppy.tests.test_ef_ph import _get_ph_base_options,\
                                     solver_available,\
                                     round_pos_sig

import mpisppy.opt.ef
import mpisppy.opt.ph

file_dir = os.path.dirname(os.path.abspath(__file__))
sizes_dir = os.path.join(file_dir,'examples','sizes')

class Test_sizes_abstract(unittest.TestCase):
    """ Test PySPModel using abstract sizes case """

    def setUp(self):
        self.pysp_sizes3 = PySPModel(os.path.join(sizes_dir,
                                     'ReferenceModel.py'),
                                     os.path.join(sizes_dir,
                                     'SIZES3', 'ScenarioStructure.dat')
                                    )

    def test_ph_constructor(self):
        pysp_sizes = self.pysp_sizes3
        options = _get_ph_base_options()
        options['PHIterLimit'] = 0
        ph = mpisppy.opt.ph.PH(options,
                                  pysp_sizes.all_scenario_names,
                                  pysp_sizes.scenario_creator,
                                  lambda *args : None,
                                  )

    def test_ef_constructor(self):
        pysp_sizes = self.pysp_sizes3
        options = {"solver": "cplex"}
        ef = mpisppy.opt.ef.ExtensiveForm(
            options,
            pysp_sizes.all_scenario_names,
            pysp_sizes.scenario_creator,
        )

    
    @unittest.skipIf(not solver_available,
                     "no solver is available")
    def test_ef_solve(self):
        pysp_sizes = self.pysp_sizes3
        options = _get_ph_base_options()
        options = {"solver": options["solvername"]}
        ef = mpisppy.opt.ef.ExtensiveForm(
            options,
            pysp_sizes.all_scenario_names,
            pysp_sizes.scenario_creator,
        )
        results = ef.solve_extensive_form(tee=False)
        sig2eobj = round_pos_sig(pyo.value(ef.ef.EF_Obj),2)
        self.assertEqual(220000.0, sig2eobj)


if __name__ == '__main__':
    unittest.main()



# Copyright 2020 by B. Knueven, D. Mildebrath, C. Muir, J-P Watson, and D.L. Woodruff
# This software is distributed under the 3-clause BSD License.
# PH-specific code

import mpisppy.phbase
import shutil
import mpi4py.MPI as mpi

# decorator snarfed from stack overflow - allows per-rank profile output file generation.
def profile(filename=None, comm=mpi.COMM_WORLD):
    pass

fullcomm = mpi.COMM_WORLD
global_rank = fullcomm.Get_rank()


############################################################################
class PH(mpisppy.phbase.PHBase):
    """ PH. See PHBase for list of args. """

    #======================================================================
    # uncomment the line below to get per-rank profile outputs, which can 
    # be examined with snakeviz (or your favorite profile output analyzer)
    #@profile(filename="profile_out")
    def ph_main(self, finalize=True):
        """ Execute the PH algorithm.

        Args:
            finalize (bool, optional, default=True):
                If True, call `PH.post_loops()`, if False, do not,
                and return None for Eobj

        Returns:
            tuple:
                Tuple containing

                conv (float):
                    The convergence value (not easily interpretable).
                Eobj (float or `None`):
                    If `finalize=True`, this is the expected, weighted
                    objective value with the proximal term included. This value
                    is not directly useful. If `finalize=False`, this value is
                    `None`.
                trivial_bound (float):
                    The "trivial bound", computed by solving the model with no
                    nonanticipativity constraints (immediately after iter 0).

        NOTE:
            You need an xhat finder either in denoument or in an extension.
        """
        verbose = self.options['verbose']
        self.PH_Prep()
        # Why is subproblem_creation() not called in PH_Prep? Answer: xhat_eval.
        self.subproblem_creation(verbose)

        if (verbose):
            print('Calling PH Iter0 on global rank {}'.format(global_rank))
        trivial_bound = self.Iter0()
        if (verbose):
            print ('Completed PH Iter0 on global rank {}'.format(global_rank))
        if ('asynchronousPH' in self.options) and (self.options['asynchronousPH']):
            raise RuntimeError("asynchronousPH is deprecated; use APH")

        self.iterk_loop()

        if finalize:
            Eobj = self.post_loops(self.extensions)
        else:
            Eobj = None

        return self.conv, Eobj, trivial_bound


if __name__ == "__main__":
    #==============================
    # hardwired by dlw for debugging
    import mpisppy.tests.examples.farmer as refmodel
    import mpisppy.utils.sputils as sputils

    PHopt = {}
    PHopt["asynchronousPH"] = False
    PHopt["solvername"] = "cplex"
    PHopt["PHIterLimit"] = 5
    PHopt["defaultPHrho"] = 1
    PHopt["convthresh"] = 0.001
    PHopt["verbose"] = True
    PHopt["display_timing"] = True
    PHopt["display_progress"] = True
    # one way to set up options (never mind that this is not a MIP)
    PHopt["iter0_solver_options"]\
        = sputils.option_string_to_dict("mipgap=0.01")
    # another way
    PHopt["iterk_solver_options"] = {"mipgap": 0.001}

    ScenCount = 50
    all_scenario_names = ['scen' + str(i) for i in range(ScenCount)]
    # end hardwire

    scenario_creator = refmodel.scenario_creator
    scenario_denouement = refmodel.scenario_denouement

    # now test extensions and convergers together.
    # NOTE that the options is a reference,
    #   so you could have changed PHopt instead.
    PHopt["PHIterLimit"] = 5 # the converger will probably stop it.
    PHopt["asynchronousPH"] = False

    from mpisppy.extensions.xhatlooper import XhatLooper
    from mpisppy.convergers.fracintsnotconv import FractionalConverger
    PHopt["xhat_looper_options"] =  {"xhat_solver_options":\
                                     PHopt["iterk_solver_options"],
                                     "scen_limit": 3,
                                     "csvname": "looper.csv"}
    ph = PH(PHopt, all_scenario_names, scenario_creator, scenario_denouement,
            extensions=XhatLooper,
            PH_converger=FractionalConverger,
            rho_setter=None)
    conv, obj, bnd = ph.ph_main()
    print ("Quitting Early.")
    quit()

    from mpisppy.extensions.diagnoser import Diagnoser
    # now test whatever is new
    ph = PH(PHopt, all_scenario_names, scenario_creator, scenario_denouement,
            extensions=Diagnoser, 
            PH_converger=None,
            rho_setter=None)
    ph.options["PHIterLimit"] = 3

    if global_rank == 0:
        try:
            shutil.rmtree("delme_diagdir")
            print ("...deleted delme_diagdir")
        except:
            pass
    ph.options["diagnoser_options"] = {"diagnoser_outdir": "delme_diagdir"}
    conv, obj, bnd = ph.ph_main()

    import mpisppy.extensions.avgminmaxer as minmax_extension
    from mpisppy.extensions.avgminmaxer import MinMaxAvg
    ph = PH(PHopt, all_scenario_names, scenario_creator, scenario_denouement,
            extensions=MinMaxAvg,
            PH_converger=None,
            rho_setter=None)
    ph.options["avgminmax_name"] =  "FirstStageCost"
    ph.options["PHIterLimit"] = 3
    conv, obj, bnd = ph.ph_main()

        # trying asynchronous operation
    PHopt["asynchronousPH"] = True
    PHopt["async_frac_needed"] = 0.5
    PHopt["async_sleep_secs"] = 1
    ph = PH(PHopt, all_scenario_names, scenario_creator, scenario_denouement)

    conv, obj, bnd = ph.ph_main()

    if global_rank == 0:
        print ("E[obj] for converged solution (probably NOT non-anticipative)",
               obj)

    dopts = sputils.option_string_to_dict("mipgap=0.001")
    objbound = ph.post_solve_bound(solver_options=dopts, verbose=False)
    if global_rank == 0:
        print ("**** Lagrangian objective function bound=",objbound)
        print ("(probably converged way too early, BTW)")
   



# Copyright 2020 by B. Knueven, D. Mildebrath, C. Muir, J-P Watson, and D.L. Woodruff
# This software is distributed under the 3-clause BSD License.
# set up the most common parser args for mpi-sppy examples
""" NOTE TO NEW USERS: just using these parsers will not, itself, do anything.
    You have to use the values when you create the dictionaries that are passed
    to spin_the_wheel. Further note that not all examples use all values
    in the parsers.
"""
import argparse

def _common_args(inparser):
    # NOTE: if you want abbreviations, override the arguments in your example
    # do not add abbreviations here.

    parser = inparser
    parser.add_argument("--max-iterations",
                        help="ph max-iterations (default 1)",
                        dest="max_iterations",
                        type=int,
                        default=1)

    parser.add_argument("--solver-name",
                        help = "solver name (default gurobi)",
                        dest="solver_name",
                        type = str,
                        default="gurobi")

    parser.add_argument("--seed",
                        help="Seed for random numbers (default is 1134)",
                        dest="seed",
                        type=int,
                        default=1134)

    parser.add_argument("--default-rho",
                        help="Global rho for PH (default None)",
                        dest="default_rho",
                        type=float,
                        default=None)

    parser.add_argument("--bundles-per-rank",
                        help="bundles per rank (default 0 (no bundles))",
                        dest="bundles_per_rank",
                        type=int,
                        default=0)                

    parser.add_argument('--with-verbose',
                        help="verbose output",
                        dest='with_verbose',
                        action='store_true')
    parser.add_argument('--no-verbose',
                        help="do not verbose output (default)",
                        dest='with_verbose',
                        action='store_false')
    parser.set_defaults(with_verbose=False)

    parser.add_argument('--with-display-progress',
                        help="display progress at each iteration",
                        dest='with_display_progress',
                        action='store_true')
    parser.add_argument('--no-display-progress',
                        help="do not display progress at each iteration (default)",
                        dest='with_display_progress',
                        action='store_false')
    parser.set_defaults(with_display_progress=False)

    parser.add_argument('--with-display-convergence-detail',
                        help="display non-anticipative variable convergence statistics at each iteration",
                        dest='with_display_convergence_detail',
                        action='store_true')
    parser.add_argument('--no-display-convergence-detail',
                        help="do not display non-anticipative variable convergence statistics at each iteration (default)",
                        dest='with_display_convergence_detail',
                        action='store_false')
    parser.set_defaults(with_display_convergence_detail=False)    

    parser.add_argument("--max-solver-threads",
                        help="Limit on threads per solver (default None)",
                        dest="max_solver_threads",
                        type=int,
                        default=None)

    parser.add_argument("--intra-hub-conv-thresh",
                        help="Within hub convergence threshold (default 1e-10)",
                        dest="intra_hub_conv_thresh",
                        type=float,
                        default=1e-10)

    parser.add_argument("--trace-prefix",
                        help="Prefix for bound spoke trace files. If None "
                             "bound spoke trace files are not written.",
                        dest="trace_prefix",
                        type=str,
                        default=None)

    parser.add_argument("--with-tee-rank0-solves",
                        help="Some cylinders support tee of rank 0 solves."
                        "(With multiple cylinder this could be confusing.)",
                        dest="tee_rank0_solves",
                        action='store_true')
    parser.add_argument("--no-tee-rank0-solves",
                        help="Some cylinders support tee of rank 0 solves.",
                        dest="tee_rank0_solves",
                        action='store_false')
    parser.set_defaults(tee_rank0_solves=False)

    parser.add_argument("--auxilliary",
                        help="Free text for use by hackers (default '').",
                        dest="auxilliary",
                        type=str,
                        default='')

    parser.add_argument("--linearize-binary-proximal-terms",
                        help="For PH, linearize the proximal terms for "
                        "all binary nonanticipative variables",
                        dest="linearize_binary_proximal_terms",
                        action='store_true')

    parser.add_argument("--linearize-proximal-terms",
                        help="For PH, linearize the proximal terms for "
                        "all nonanticipative variables",
                        dest="linearize_proximal_terms",
                        action='store_true')

    parser.add_argument("--proximal-linearization-tolerance",
                        help="For PH, when linearizing proximal terms, "
                        "a cut will be added if the proximal term approximation "
                        "is looser than this value (default 1e-1)",
                        dest="proximal_linearization_tolerance",
                        type=float,
                        default=1.e-1)

    return parser
    
def make_parser(progname=None, num_scens_reqd=False):
    # make a parser for the program named progname
    # NOTE: if you want abbreviations, override the arguments in your example
    # do not add abbreviations here.
    parser = argparse.ArgumentParser(prog=progname, conflict_handler="resolve")

    if num_scens_reqd:
        parser.add_argument(
            "num_scens", help="Number of scenarios", type=int
        )
    else:
        parser.add_argument(
            "--num-scens",
            help="Number of scenarios (default None)",
            dest="num_scens",
            type=int,
            default=None,
        )
    parser = _common_args(parser)
    return parser

def _basic_multistage(progname=None):
    parser = argparse.ArgumentParser(prog=progname, conflict_handler="resolve")

    # the default is intended more as an example than as a default
    parser.add_argument("--BFs",
                        help="Comma delimied branching factors (default 2,2)",
                        dest="BFs",
                        type=str,
                        default="2,2")
    return parser


def make_multistage_parser(progname=None):
    # make a parser for the program named progname
    # NOTE: if you want abbreviations, override the arguments in your example
    # do not add abbreviations here.
    parser = _basic_multistage(progname=None)
    parser = _common_args(parser)
    return parser

#### EF ####
def make_EF2_parser(progname=None, num_scens_reqd=False):
    # create a parser just for EF two-stage (does not call _common_args)
    # NOTE: if you want abbreviations, override the arguments in your example
    # do not add abbreviations here.
    parser = argparse.ArgumentParser(prog=progname, conflict_handler="resolve")

    if num_scens_reqd:
        parser.add_argument(
            "num_scens", help="Number of scenarios", type=int
        )
    else:
        parser.add_argument(
            "--num-scens",
            help="Number of scenarios (default None)",
            dest="num_scens",
            type=int,
            default=None,
        )
        
    parser.add_argument("--EF-solver-name",
                        help = "solver name (default gurobi)",
                        dest="EF_solver_name",
                        type = str,
                        default="gurobi")


    parser.add_argument("--EF-mipgap",
                        help="mip gap option for the solver if needed (default None)",
                        dest="EF_mipgap",
                        type=float,
                        default=None)
    return parser

def make_EF_multistage_parser(progname=None, num_scens_reqd=False):
    # create a parser just for EF two-stage (does not call _common_args)
    # NOTE: if you want abbreviations, override the arguments in your example
    # do not add abbreviations here.
    parser = _basic_multistage(progname=None)
    parser.add_argument("--EF-solver-name",
                        help = "solver name (default gurobi)",
                        dest="EF_solver_name",
                        type = str,
                        default="gurobi")


    parser.add_argument("--EF-mipgap",
                        help="mip gap option for the solver if needed (default None)",
                        dest="EF_mipgap",
                        type=float,
                        default=None)
    return parser
##### common additions to the command line #####

def two_sided_args(inparser):
    # add commands to inparser and also return the result
    parser = inparser
    parser.add_argument("--rel-gap",
                        help="relative termination gap (default 0.05)",
                        dest="rel_gap",
                        type=float,
                        default=0.05)

    parser.add_argument("--abs-gap",
                        help="absolute termination gap (default 8)",
                        dest="abs_gap",
                        type=float,
                        default=8.)

    return parser

def mip_options(inparser):
    parser = inparser
    parser.add_argument("--iter0-mipgap",
                        help="mip gap option for iteration 0 (default None)",
                        dest="iter0_mipgap",
                        type=float,
                        default=None)

    parser.add_argument("--iterk-mipgap",
                        help="mip gap option non-zero iterations (default None)",
                        dest="iterk_mipgap",
                        type=float,
                        default=None)
    return parser

def aph_args(inparser):
    parser = inparser
    parser.add_argument('--aph-gamma',
                        help='Gamma parameter associated with asychronous projective hedging (default 1.0)',
                        dest='aph_gamma',
                        type=float,
                        default=1.0)
    parser.add_argument('--aph-nu',
                        help='Nu parameter associated with asychronous projective hedging (default 1.0)',
                        dest="aph_nu",
                        type=float,
                        default=1.0)
    parser.add_argument('--aph-frac-needed',
                        help='Fraction of sub-problems required before computing projective step (default 1.0)',
                        dest='aph_frac_needed',
                        type=float,
                        default=1.0)
    parser.add_argument('--aph-dispatch-frac',
                        help='Fraction of sub-problems to dispatch at each step of asychronous projective hedging (default 1.0)',
                        dest='aph_dispatch_frac',
                        type=float,
                        default=1.0)
    parser.add_argument('--aph-sleep-seconds',
                        help='Spin-lock sleep time for APH (default 0.01)',
                        dest='aph_sleep_seconds',
                        type=float,
                        default=0.01)    
    return parser    

def fixer_args(inparser):
    parser = inparser
    parser.add_argument('--with-fixer',
                        help="have an integer fixer extension (default)",
                        dest='with_fixer',
                        action='store_true')
    parser.add_argument('--no-fixer',
                        help="do not have an integer fixer extension",
                        dest='with_fixer',
                        action='store_false')
    parser.set_defaults(with_fixer=True)

    parser.add_argument("--fixer-tol",
                        help="fixer bounds tolerance  (default 1e-4)",
                        dest="fixer_tol",
                        type=float,
                        default=1e-2)
    return parser


def fwph_args(inparser):
    parser = inparser
    parser.add_argument('--with-fwph',
                        help="have an fwph spoke (default)",
                        dest='with_fwph',
                        action='store_true')
    parser.add_argument('--no-fwph',
                        help="do not have an fwph spoke",
                        dest='with_fwph',
                        action='store_false')
    parser.set_defaults(with_fwph=True)

    parser.add_argument("--fwph-iter-limit",
                        help="maximum fwph iterations (default 10)",
                        dest="fwph_iter_limit",
                        type=int,
                        default=10)

    parser.add_argument("--fwph-weight",
                        help="fwph weight (default 0)",
                        dest="fwph_weight",
                        type=float,
                        default=0.0)

    parser.add_argument("--fwph-conv-thresh",
                        help="fwph convergence threshold  (default 1e-4)",
                        dest="fwph_conv_thresh",
                        type=float,
                        default=1e-4)

    parser.add_argument("--fwph-stop-check-tol",
                        help="fwph tolerance for Gamma^t (default 1e-4)",
                        dest="fwph_stop_check_tol",
                        type=float,
                        default=1e-4)

    parser.add_argument("--fwph-mipgap",
                        help="mip gap option FW subproblems iterations (default None)",
                        dest="fwph_mipgap",
                        type=float,
                        default=None)

    return parser

def lagrangian_args(inparser):
    parser = inparser
    parser.add_argument('--with-lagrangian',
                        help="have a lagrangian spoke (default)",
                        dest='with_lagrangian',
                        action='store_true')
    parser.add_argument('--no-lagrangian',
                        help="do not have a lagrangian spoke",
                        dest='with_lagrangian',
                        action='store_false')
    parser.set_defaults(with_lagrangian=True)

    parser.add_argument("--lagrangian-iter0-mipgap",
                        help="lgr. iter0 solver option mipgap (default None)",
                        dest="lagrangian_iter0_mipgap",
                        type=float,
                        default=None)

    parser.add_argument("--lagrangian-iterk-mipgap",
                        help="lgr. iterk solver option mipgap (default None)",
                        dest="lagrangian_iterk_mipgap",
                        type=float,
                        default=None)

    return parser


def lagranger_args(inparser):
    parser = inparser
    parser.add_argument('--with-lagranger',
                        help="have a special lagranger spoke (default)",
                        dest='with_lagranger',
                        action='store_true')
    parser.add_argument('--no-lagranger',
                        help="do not have a special lagranger spoke",
                        dest='with_lagranger',
                        action='store_false')
    parser.set_defaults(with_lagranger=True)

    parser.add_argument("--lagranger-iter0-mipgap",
                        help="lagranger iter0 mipgap (default None)",
                        dest="lagranger_iter0_mipgap",
                        type=float,
                        default=None)

    parser.add_argument("--lagranger-iterk-mipgap",
                        help="lagranger iterk mipgap (default None)",
                        dest="lagranger_iterk_mipgap",
                        type=float,
                        default=None)

    parser.add_argument("--lagranger-rho-rescale-factors-json",
                        help="json file: rho rescale factors (default None)",
                        dest="lagranger_rho_rescale_factors_json",
                        type=float,
                        default=None)

    return parser


def xhatlooper_args(inparser):
    parser = inparser
    parser.add_argument('--with-xhatlooper',
                        help="have an xhatlooper spoke",
                        dest='with_xhatlooper',
                        action='store_true')
    parser.add_argument('--no-xhatlooper',
                        help="do not have an xhatlooper spoke (default)",
                        dest='with_xhatlooper',
                        action='store_false')
    parser.set_defaults(with_xhatlooper=False)
    parser.add_argument("--xhat-scen-limit",
                        help="scenario limit xhat looper to try (default 3)",
                        dest="xhat_scen_limit",
                        type=int,
                        default=3)

    return parser


def xhatshuffle_args(inparser):
    parser = inparser
    parser.add_argument('--with-xhatshuffle',
                        help="have an xhatshuffle spoke (default)",
                        dest='with_xhatshuffle',
                        action='store_true')
    parser.add_argument('--no-xhatshuffle',
                        help="do not have an xhatshuffle spoke",
                        dest='with_xhatshuffle',
                        action='store_false')
    parser.set_defaults(with_xhatshuffle=True)

    return parser


def xhatspecific_args(inparser):
    # we will not try to get the specification from the command line
    parser = inparser
    parser.add_argument('--with-xhatspecific',
                        help="have an xhatspecific spoke (default)",
                        dest='with_xhatspecific',
                        action='store_true')
    parser.add_argument('--no-xhatspecific',
                        help="do not have an xhatspecific spoke",
                        dest='with_xhatspecific',
                        action='store_false')
    parser.set_defaults(with_xhatspecific=True)

    return parser


def xhatlshaped_args(inparser):
    # we will not try to get the specification from the command line
    parser = inparser
    parser.add_argument('--with-xhatlshaped',
                        help="have an xhatlshaped spoke (default)",
                        dest='with_xhatlshaped',
                        action='store_true')
    parser.add_argument('--no-xhatlshaped',
                        help="do not have an xhatlshaped spoke",
                        dest='with_xhatlshaped',
                        action='store_false')
    parser.set_defaults(with_xhatlshaped=True)

    return parser

def slamup_args(inparser):
    # we will not try to get the specification from the command line
    parser = inparser
    parser.add_argument('--with-slamup',
                        help="have an slamup spoke (default)",
                        dest='with_slamup',
                        action='store_true')
    parser.add_argument('--no-slamup',
                        help="do not have an slamup spoke",
                        dest='with_slamup',
                        action='store_false')
    parser.set_defaults(with_slamup=True)

    return parser

def slamdown_args(inparser):
    # we will not try to get the specification from the command line
    parser = inparser
    parser.add_argument('--with-slamdown',
                        help="have an slamdown spoke (default)",
                        dest='with_slamdown',
                        action='store_true')
    parser.add_argument('--no-slamdown',
                        help="do not have an slamdown spoke",
                        dest='with_slamdown',
                        action='store_false')
    parser.set_defaults(with_slamdown=True)

    return parser

def cross_scenario_cuts_args(inparser):
    # we will not try to get the specification from the command line
    parser = inparser
    parser.add_argument('--with-cross-scenario-cuts',
                        help="have a cross scenario cut spoke (default)",
                        dest='with_cross_scenario_cuts',
                        action='store_true')
    parser.add_argument('--no-cross-scenario-cuts',
                        help="do not have a cross scenario cut spoke",
                        dest='with_cross_scenario_cuts',
                        action='store_false')
    parser.set_defaults(with_cross_scenario_cuts=True)

    parser.add_argument("--cross-scenario-iter-cnt",
                        help="cross scen check bound improve iterations "
                        "(default 4)",
                        dest="cross_scenario_iter_cnt",
                        type=int,
                        default=4)

    parser.add_argument("--eta-bounds-mipgap",
                        help="mipgap for determining eta bounds for cross "
                        "scenario cuts (default 0.01)",
                        dest="eta_bounds_mipgap",
                        type=float,
                        default=0.01)

    return parser



# Copyright 2020 by B. Knueven, D. Mildebrath, C. Muir, J-P Watson, and D.L. Woodruff
# This software is distributed under the 3-clause BSD License.
# An extension to output diagnostic information at each iteration
# to a file for each scenario.
# DLW, March 2019
# This extension uses options["diagnostics_outdir"]
# This could be used as a starting point for writing your own diangostics;
#   for "canned" diagnostics, use the baseparser PH option with-display-convergence-detail

import datetime as dt
import os
import pyomo.environ as pyo
import mpisppy.extensions.xhatbase
from mpisppy.utils.sputils import find_active_objective

class Diagnoser(mpisppy.extensions.xhatbase.XhatBase):
    """
    Args:
        ph (PH object): the calling object
        rank (int): mpi process rank of currently running process
    """
    def __init__(self, ph):
        dirname = ph.options["diagnoser_options"]["diagnoser_outdir"]
        if os.path.exists(dirname):
            if ph.cylinder_rank == 0:
                print ("Shutting down because Diagnostic directory exists:",
                       dirname)
            quit()
        if ph.cylinder_rank == 0:
            os.mkdir(dirname) # just let it crash

        super().__init__(ph)
        self.options = self.ph.options["diagnoser_options"]
        self.dirname = self.options["diagnoser_outdir"]

    def write_loop(self):
        """ Bundles are special. Also: this code needs help
        from the ph object to be more efficient...
        """
        for sname, s in self.ph.local_scenarios.items():
            bundling = self.ph.bundling
            fname = self.dirname+os.sep+sname+".dag"
            with open(fname, "a") as f:
                f.write(str(self.ph._PHIter)+",")
                if not bundling:
                    objfct = find_active_objective(s)
                    f.write(str(pyo.value(objfct)))
                else:
                    f.write("Bundling"+",")
                    f.write(str(pyo.value(self.ph.saved_objs[sname])))
                f.write("\n")

    def pre_iter0(self):
        return

    def post_iter0(self):
        for sname, s in self.ph.local_scenarios.items():
            fname = self.dirname+os.sep+sname+".dag"
            with open(fname, "w") as f:
                f.write(str(dt.datetime.now())+", diagnoser\n")

        self.write_loop()
        
    def miditer(self, PHIter, conv):
        return

    def enditer(self, PHIter):
        self.write_loop()

    def post_everything(self, PHIter, conv):
        return



# special for ph debugging DLW Dec 2018
# unlimited crops
# ALL INDEXES ARE ZERO-BASED
#  ___________________________________________________________________________
#
#  Pyomo: Python Optimization Modeling Objects
#  Copyright 2018 National Technology and Engineering Solutions of Sandia, LLC
#  Under the terms of Contract DE-NA0003525 with National Technology and 
#  Engineering Solutions of Sandia, LLC, the U.S. Government retains certain 
#  rights in this software.
#  This software is distributed under the 3-clause BSD License.
#  ___________________________________________________________________________
#
# special scalable farmer for stress-testing

import pyomo.environ as pyo
import numpy as np
import mpisppy.scenario_tree as scenario_tree
import mpisppy.utils.sputils as sputils

# Use this random stream:
farmerstream = np.random.RandomState()

def scenario_creator(
    scenario_name, use_integer=False, sense=pyo.minimize, crops_multiplier=1,
    num_scens=None
):
    """ Create a scenario for the (scalable) farmer example.
    
    Args:
        scenario_name (str):
            Name of the scenario to construct.
        use_integer (bool, optional):
            If True, restricts variables to be integer. Default is False.
        sense (int, optional):
            Model sense (minimization or maximization). Must be either
            pyo.minimize or pyo.maximize. Default is pyo.minimize.
        crops_multiplier (int, optional):
            Factor to control scaling. There will be three times this many
            crops. Default is 1.
        num_scens (int, optional):
            Number of scenarios. We use it to compute _mpisppy_probability. 
            Default is None.
    """
    # scenario_name has the form <str><int> e.g. scen12, foobar7
    # The digits are scraped off the right of scenario_name using regex then
    # converted mod 3 into one of the below avg./avg./above avg. scenarios
    scennum   = sputils.extract_num(scenario_name)
    basenames = ['BelowAverageScenario', 'AverageScenario', 'AboveAverageScenario']
    basenum   = scennum  % 3
    groupnum  = scennum // 3
    scenname  = basenames[basenum]+str(groupnum)

    # The RNG is seeded with the scenario number so that it is
    # reproducible when used with multiple threads.
    # NOTE: if you want to do replicates, you will need to pass a seed
    # as a kwarg to scenario_creator then use seed+scennum as the seed argument.
    farmerstream.seed(scennum)

    # Check for minimization vs. maximization
    if sense not in [pyo.minimize, pyo.maximize]:
        raise ValueError("Model sense Not recognized")

    # Create the concrete model object
    model = pysp_instance_creation_callback(
        scenname,
        use_integer=use_integer,
        sense=sense,
        crops_multiplier=crops_multiplier,
    )

    # Create the list of nodes associated with the scenario (for two stage,
    # there is only one node associated with the scenario--leaf nodes are
    # ignored).
    model._mpisppy_node_list = [
        scenario_tree.ScenarioNode(
            name="ROOT",
            cond_prob=1.0,
            stage=1,
            cost_expression=model.FirstStageCost,
            scen_name_list=None, # Deprecated?
            nonant_list=[model.DevotedAcreage],
            scen_model=model,
        )
    ]
    
    #Add the probability of the scenario
    if num_scens is not None :
        model._mpisppy_probability = 1/num_scens
    return model

def pysp_instance_creation_callback(
    scenario_name, use_integer=False, sense=pyo.minimize, crops_multiplier=1
):
    # long function to create the entire model
    # scenario_name is a string (e.g. AboveAverageScenario0)
    #
    # Returns a concrete model for the specified scenario

    # scenarios come in groups of three
    scengroupnum = sputils.extract_num(scenario_name)
    scenario_base_name = scenario_name.rstrip("0123456789")
    
    model = pyo.ConcreteModel()

    def crops_init(m):
        retval = []
        for i in range(crops_multiplier):
            retval.append("WHEAT"+str(i))
            retval.append("CORN"+str(i))
            retval.append("SUGAR_BEETS"+str(i))
        return retval

    model.CROPS = pyo.Set(initialize=crops_init)

    #
    # Parameters
    #

    model.TOTAL_ACREAGE = 500.0 * crops_multiplier

    def _scale_up_data(indict):
        outdict = {}
        for i in range(crops_multiplier):
           for crop in ['WHEAT', 'CORN', 'SUGAR_BEETS']:
               outdict[crop+str(i)] = indict[crop]
        return outdict
        
    model.PriceQuota = _scale_up_data(
        {'WHEAT':100000.0,'CORN':100000.0,'SUGAR_BEETS':6000.0})

    model.SubQuotaSellingPrice = _scale_up_data(
        {'WHEAT':170.0,'CORN':150.0,'SUGAR_BEETS':36.0})

    model.SuperQuotaSellingPrice = _scale_up_data(
        {'WHEAT':0.0,'CORN':0.0,'SUGAR_BEETS':10.0})

    model.CattleFeedRequirement = _scale_up_data(
        {'WHEAT':200.0,'CORN':240.0,'SUGAR_BEETS':0.0})

    model.PurchasePrice = _scale_up_data(
        {'WHEAT':238.0,'CORN':210.0,'SUGAR_BEETS':100000.0})

    model.PlantingCostPerAcre = _scale_up_data(
        {'WHEAT':150.0,'CORN':230.0,'SUGAR_BEETS':260.0})

    #
    # Stochastic Data
    #
    Yield = {}
    Yield['BelowAverageScenario'] = \
        {'WHEAT':2.0,'CORN':2.4,'SUGAR_BEETS':16.0}
    Yield['AverageScenario'] = \
        {'WHEAT':2.5,'CORN':3.0,'SUGAR_BEETS':20.0}
    Yield['AboveAverageScenario'] = \
        {'WHEAT':3.0,'CORN':3.6,'SUGAR_BEETS':24.0}

    def Yield_init(m, cropname):
        # yield as in "crop yield"
        crop_base_name = cropname.rstrip("0123456789")
        if scengroupnum != 0:
            return Yield[scenario_base_name][crop_base_name]+farmerstream.rand()
        else:
            return Yield[scenario_base_name][crop_base_name]

    model.Yield = pyo.Param(model.CROPS,
                            within=pyo.NonNegativeReals,
                            initialize=Yield_init,
                            mutable=True)

    #
    # Variables
    #

    if (use_integer):
        model.DevotedAcreage = pyo.Var(model.CROPS,
                                       within=pyo.NonNegativeIntegers,
                                       bounds=(0.0, model.TOTAL_ACREAGE))
    else:
        model.DevotedAcreage = pyo.Var(model.CROPS, 
                                       bounds=(0.0, model.TOTAL_ACREAGE))

    model.QuantitySubQuotaSold = pyo.Var(model.CROPS, bounds=(0.0, None))
    model.QuantitySuperQuotaSold = pyo.Var(model.CROPS, bounds=(0.0, None))
    model.QuantityPurchased = pyo.Var(model.CROPS, bounds=(0.0, None))

    #
    # Constraints
    #

    def ConstrainTotalAcreage_rule(model):
        return pyo.sum_product(model.DevotedAcreage) <= model.TOTAL_ACREAGE

    model.ConstrainTotalAcreage = pyo.Constraint(rule=ConstrainTotalAcreage_rule)

    def EnforceCattleFeedRequirement_rule(model, i):
        return model.CattleFeedRequirement[i] <= (model.Yield[i] * model.DevotedAcreage[i]) + model.QuantityPurchased[i] - model.QuantitySubQuotaSold[i] - model.QuantitySuperQuotaSold[i]

    model.EnforceCattleFeedRequirement = pyo.Constraint(model.CROPS, rule=EnforceCattleFeedRequirement_rule)

    def LimitAmountSold_rule(model, i):
        return model.QuantitySubQuotaSold[i] + model.QuantitySuperQuotaSold[i] - (model.Yield[i] * model.DevotedAcreage[i]) <= 0.0

    model.LimitAmountSold = pyo.Constraint(model.CROPS, rule=LimitAmountSold_rule)

    def EnforceQuotas_rule(model, i):
        return (0.0, model.QuantitySubQuotaSold[i], model.PriceQuota[i])

    model.EnforceQuotas = pyo.Constraint(model.CROPS, rule=EnforceQuotas_rule)

    # Stage-specific cost computations;

    def ComputeFirstStageCost_rule(model):
        return pyo.sum_product(model.PlantingCostPerAcre, model.DevotedAcreage)
    model.FirstStageCost = pyo.Expression(rule=ComputeFirstStageCost_rule)

    def ComputeSecondStageCost_rule(model):
        expr = pyo.sum_product(model.PurchasePrice, model.QuantityPurchased)
        expr -= pyo.sum_product(model.SubQuotaSellingPrice, model.QuantitySubQuotaSold)
        expr -= pyo.sum_product(model.SuperQuotaSellingPrice, model.QuantitySuperQuotaSold)
        return expr
    model.SecondStageCost = pyo.Expression(rule=ComputeSecondStageCost_rule)

    def total_cost_rule(model):
        if (sense == pyo.minimize):
            return model.FirstStageCost + model.SecondStageCost
        return -model.FirstStageCost - model.SecondStageCost
    model.Total_Cost_Objective = pyo.Objective(rule=total_cost_rule, 
                                               sense=sense)

    return model


#=========
def scenario_names_creator(num_scens,start=None):
    # (only for Amalgomator): return the full list of num_scens scenario names
    # if start!=None, the list starts with the 'start' labeled scenario
    if (start is None) :
        start=0
    return [f"scen{i}" for i in range(start,start+num_scens)]
        


#=========
def inparser_adder(inparser):
    # (only for Amalgomator): add command options unique to farmer
    inparser.add_argument("--crops-multiplier",
                          help="number of crops will be three times this (default 1)",
                          dest="crops_multiplier",
                          type=int,
                          default=1)
    
    inparser.add_argument("--farmer-with-integers",
                          help="make the version that has integers (default False)",
                          dest="use_integer",
                          action="store_true",)
    inparser.set_defaults(use_integer=False)


#=========
def kw_creator(options):
    # (only for Amalgomator): linked to the scenario_creator and inparser_adder
    kwargs = {"use_integer": options['use_integer'] if 'use_integer' in options else False,
              "crops_multiplier": options['crops_multiplier'] if 'crops_multiplier' in options else 1,
              "num_scens" : options['num_scens'] if 'num_scens' in options else None,
              }
    return kwargs


#============================
def scenario_denouement(rank, scenario_name, scenario):
    sname = scenario_name
    s = scenario
    if sname == 'scen0':
        print("Arbitrary sanity checks:")
        print ("SUGAR_BEETS0 for scenario",sname,"is",
               pyo.value(s.DevotedAcreage["SUGAR_BEETS0"]))
        print ("FirstStageCost for scenario",sname,"is", pyo.value(s.FirstStageCost))
        



# This software is distributed under the 3-clause BSD License.
#!/bin/usr/env python3
import glob
import sys
import os

# We raise an error if trying to install with python2
if sys.version[0] == '2':
    print("Error: This package must be installed with python3")
    sys.exit(1)

from setuptools import find_packages
from distutils.core import setup

packages = find_packages()

setup(
    name='mpi-sppy',
    version='0.9',
    description="mpi-sppy",
    url='https://github.com/Pyomo/mpi-sppy',
    author='David Woodruff',
    author_email='dlwoodruff@ucdavis.edu',
    packages=packages,
    install_requires=[
        'numpy>=1.19',
        'pyomo>=6.0'
    ]
)



# Copyright 2020 by B. Knueven, D. Mildebrath, C. Muir, J-P Watson, and D.L. Woodruff
# This software is distributed under the 3-clause BSD License.
# base class for hub and for spoke strata

import os
import time
import logging
import weakref
import numpy as np
import pyomo.environ as pyo
import mpisppy.utils.sputils as sputils

from mpi4py import MPI

from mpisppy import global_toc

logger = logging.getLogger("SPBase")
logger.setLevel(logging.WARN)


class SPBase:
    """ Defines an interface to all strata (hubs and spokes)

        Args:
            options (dict): options
            all_scenario_names (list): all scenario names
            scenario_creator (fct): returns a concrete model with special things
            scenario_denouement (fct): for post processing and reporting
            all_nodenames (list): all non-leaf node names; can be None for 2 Stage
            mpicomm (MPI comm): if not given, use the global fullcomm
            scenario_creator_kwargs (dict): kwargs passed directly to
                scenario_creator.
            variable_probability (fct): returns a list of tuples of (id(var), prob)
                to set variable-specific probability (similar to PHBase.rho_setter).

        Attributes:
          local_scenarios (dict of scenario objects): concrete models with 
                extra data, key is name
          comms (dict): keys are node names values are comm objects.
          local_scenario_names (list): names of locals 
    """

    def __init__(
            self,
            options,
            all_scenario_names,
            scenario_creator,
            scenario_denouement=None,
            all_nodenames=None,
            mpicomm=None,
            scenario_creator_kwargs=None,
            variable_probability=None,
            E1_tolerance=1e-5
    ):
        # TODO add missing and private attributes (JP)
        # TODO add a class attribute called ROOTNODENAME = "ROOT"
        # TODO? add decorators to the class attributes

        self.start_time = time.perf_counter()
        self.options = options
        self.all_scenario_names = all_scenario_names
        self.scenario_creator = scenario_creator
        self.scenario_denouement = scenario_denouement
        self.comms = dict()
        self.local_scenarios = dict()
        self.local_scenario_names = list()
        self.E1_tolerance = E1_tolerance  # probs must sum to almost 1
        self.names_in_bundles = None
        self.scenarios_constructed = False
        if all_nodenames is None:
            self.all_nodenames = ["ROOT"]
        elif "ROOT" in all_nodenames:
            self.all_nodenames = all_nodenames
        else:
            raise RuntimeError("'ROOT' must be in the list of node names")
        self.variable_probability = variable_probability
        self.multistage = (len(self.all_nodenames) > 1)

        # Set up MPI communicator and rank
        if mpicomm is not None:
            self.mpicomm = mpicomm
        else:
            self.mpicomm = MPI.COMM_WORLD
        self.cylinder_rank = self.mpicomm.Get_rank()
        self.n_proc = self.mpicomm.Get_size()
        self.global_rank = MPI.COMM_WORLD.Get_rank()

        global_toc("Initializing SPBase")

        if self.n_proc > len(self.all_scenario_names):
            raise RuntimeError("More ranks than scenarios")

        # Call various initialization methods
        if "branching_factors" in self.options:
            self.branching_factors = self.options["branching_factors"]
        else:
            self.branching_factors = [len(self.all_scenario_names)]
        self._calculate_scenario_ranks()
        if "bundles_per_rank" in self.options and self.options["bundles_per_rank"] > 0:
            self._assign_bundles()
            self.bundling = True
        else:
            self.bundling = False
        self._create_scenarios(scenario_creator_kwargs)
        self._look_and_leap()
        self._compute_unconditional_node_probabilities()
        self._attach_nlens()
        self._attach_nonant_indices()
        self._attach_varid_to_nonant_index()
        self._create_communicators()
        self._verify_nonant_lengths()
        self._set_sense()
        self._use_variable_probability_setter()

        ## SPCommunicator object
        self._spcomm = None

        # for writers, if the appropriate
        # solution is loaded into the subproblems
        self.tree_solution_available = False
        self.first_stage_solution_available = False

    def _set_sense(self, comm=None):
        """ Check to confirm that all the models constructed by scenario_crator
            have the same sense (min v. max), and set self.is_minimizing
            accordingly.
        """
        is_min, clear = sputils._models_have_same_sense(self.local_scenarios)
        if not clear:
            raise RuntimeError(
                "All scenario models must have the same "
                "model sense (minimize or maximize)"
            )
        self.is_minimizing = is_min

        if self.n_proc <= 1:
            return

        # Check that all the ranks agree
        global_senses = self.mpicomm.gather(is_min, root=0)
        if self.cylinder_rank != 0:
            return
        sense = global_senses[0]
        clear = all(val == sense for val in global_senses)
        if not clear:
            raise RuntimeError(
                "All scenario models must have the same "
                "model sense (minimize or maximize)"
            )

    def _verify_nonant_lengths(self):
        local_node_nonant_lengths = {}   # keys are tree node names

        # we need to accumulate all local contributions before the reduce
        for k,s in self.local_scenarios.items():
            nlens = s._mpisppy_data.nlens
            for node in s._mpisppy_node_list:
                ndn = node.name
                mylen = nlens[ndn]
                if ndn not in local_node_nonant_lengths:
                    local_node_nonant_lengths[ndn] = mylen
                elif local_node_nonant_lengths[ndn] != mylen:
                    raise RuntimeError(f"Tree node {ndn} has different number of non-anticipative "
                            f"variables between scenarios {mylen} vs. {local_node_nonant_lengths[ndn]}")

        # compute node xbar values(reduction)
        for ndn, val in local_node_nonant_lengths.items():
            local_val = np.array([val], 'i')
            max_val = np.zeros(1, 'i')
            self.comms[ndn].Allreduce([local_val, MPI.INT],
                                      [max_val, MPI.INT],
                                      op=MPI.MAX)

            if val != int(max_val[0]):
                raise RuntimeError(f"Tree node {ndn} has different number of non-anticipative "
                        f"variables between scenarios {val} vs. max {max_val[0]}")


    def _calculate_scenario_ranks(self):
        """ Populate the following attributes
            1. self.scenario_names_to_rank (dict of dict):
                keys are comms (i.e., tree nodes); values are dicts with keys
                that are scenario names and values that are ranks within that comm

            2. self._rank_slices (list of lists)
                indices correspond to ranks in self.mpicomm and the values are a list
                of scenario indices
                rank -> list of scenario indices for that rank

            3. self._scenario_slices (list)
                indices are scenario indices and values are the rank of that scenario
                within self.mpicomm
                scenario index -> rank

            4. self._scenario_tree (instance of sputils._ScenTree)

            5. self.local_scenario_names (list)
               List of index names owned by the local rank

        """
        tree = sputils._ScenTree(self.branching_factors, self.all_scenario_names)

        self.scenario_names_to_rank, self._rank_slices, self._scenario_slices =\
                tree.scen_names_to_ranks(self.n_proc)
        self._scenario_tree = tree

        # list of scenario names owned locally
        self.local_scenario_names = [
            self.all_scenario_names[i] for i in self._rank_slices[self.cylinder_rank]
        ]


    def _assign_bundles(self):
        """ Create self.names_in_bundles, a dict of dicts
            
            self.names_in_bundles[rank number][bundle number] = 
                    list of scenarios in that bundle

        """
        scen_count = len(self.all_scenario_names)

        if self.options["verbose"] and self.cylinder_rank == 0:
            print("(rank0)", self.options["bundles_per_rank"], "bundles per rank")
        if self.n_proc * self.options["bundles_per_rank"] > scen_count:
            raise RuntimeError(
                "Not enough scenarios to satisfy the bundles_per_rank requirement"
            )

        # dict: rank number --> list of scenario names owned by rank
        names_at_rank = {
            curr_rank: [self.all_scenario_names[i] for i in slc]
            for (curr_rank, slc) in enumerate(self._rank_slices)
        }

        self.names_in_bundles = dict()
        num_bundles = self.options["bundles_per_rank"]

        for curr_rank in range(self.n_proc):
            scen_count = len(names_at_rank[curr_rank])
            avg = scen_count / num_bundles
            slices = [
                range(int(i * avg), int((i + 1) * avg)) for i in range(num_bundles)
            ]
            self.names_in_bundles[curr_rank] = {
                curr_bundle: [names_at_rank[curr_rank][i] for i in slc]
                for (curr_bundle, slc) in enumerate(slices)
            }

    def _create_scenarios(self, scenario_creator_kwargs):
        """ Call the scenario_creator for every local scenario, and store the
            results in self.local_scenarios (dict indexed by scenario names).

            Notes:
                If a scenario probability is not specified as an attribute
                _mpisppy_probability of the ConcreteModel returned by ScenarioCreator,
                this function automatically assumes uniform probabilities.
        """
        if self.scenarios_constructed:
            raise RuntimeError("Scenarios already constructed.")

        if scenario_creator_kwargs is None:
            scenario_creator_kwargs = dict()

        for sname in self.local_scenario_names:
            instance_creation_start_time = time.time()
            s = self.scenario_creator(sname, **scenario_creator_kwargs)
            self.local_scenarios[sname] = s
            if "display_timing" in self.options and self.options["display_timing"]:
                instance_creation_time = time.time() - instance_creation_start_time
                all_instance_creation_times = self.mpicomm.gather(
                    instance_creation_time, root=0
                )
                if self.cylinder_rank == 0:
                    aict = all_instance_creation_times
                    print("Scenario instance creation times:")
                    print(f"\tmin={np.min(aict):4.2f} mean={np.mean(aict):4.2f} max={np.max(aict):4.2f}")
        self.scenarios_constructed = True

    def _attach_nonant_indices(self):
        for (sname, scenario) in self.local_scenarios.items():
            _nonant_indices = dict()
            nlens = scenario._mpisppy_data.nlens        
            for node in scenario._mpisppy_node_list:
                ndn = node.name
                for i in range(nlens[ndn]):
                    _nonant_indices[ndn,i] = node.nonant_vardata_list[i]
            scenario._mpisppy_data.nonant_indices = _nonant_indices

            
    def _attach_nlens(self):
        for (sname, scenario) in self.local_scenarios.items():
            # Things need to be by node so we can bind to the
            # indices of the vardata lists for the nodes.
            scenario._mpisppy_data.nlens = {
                node.name: len(node.nonant_vardata_list)
                for node in scenario._mpisppy_node_list
            }

            # NOTE: This only is used by extensions.xhatbase.XhatBase._try_one.
            #       If that is re-factored, we can remove it here.
            scenario._mpisppy_data.cistart = dict()
            sofar = 0
            for ndn, ndn_len in scenario._mpisppy_data.nlens.items():
                scenario._mpisppy_data.cistart[ndn] = sofar
                sofar += ndn_len

                
    def _attach_varid_to_nonant_index(self):
        """ Create a map from the id of nonant variables to their Pyomo index.
        """
        for (sname, scenario) in self.local_scenarios.items():
            # In order to support rho setting, create a map
            # from the id of vardata object back its _nonant_index.
            scenario._mpisppy_data.varid_to_nonant_index =\
                {id(var): ndn_i for ndn_i, var in scenario._mpisppy_data.nonant_indices.items()}
            

    def _create_communicators(self):

        # Create communicator objects, one for each node
        nonleafnodes = dict()
        for (sname, scenario) in self.local_scenarios.items():
            for node in scenario._mpisppy_node_list:
                nonleafnodes[node.name] = node  # might be assigned&reassigned

        # check the node names given by the scenarios
        for nodename in nonleafnodes:
            if nodename not in self.all_nodenames:
                raise RuntimeError(f"Tree node '{nodename}' not in all_nodenames list {self.all_nodenames}")

        # loop over all nodes and make the comms (split requires all ranks)
        # make sure we loop in the same order, so every rank iterate over
        # the nodelist
        for nodename in self.all_nodenames:
            if nodename == "ROOT":
                self.comms["ROOT"] = self.mpicomm
            elif nodename in nonleafnodes:
                nodenumber = sputils.extract_num(nodename)
                # IMPORTANT: See note in sputils._ScenTree.scen_names_to_ranks. Need to keep
                #            this split aligned with self.scenario_names_to_rank
                self.comms[nodename] = self.mpicomm.Split(color=nodenumber, key=self.cylinder_rank)
            else: # this rank is not included in the communicator
                self.mpicomm.Split(color=MPI.UNDEFINED, key=self.n_proc)

        ## ensure we've set things up correctly for all comms
        for nodename, comm in self.comms.items():
            scenario_names_to_comm_rank = self.scenario_names_to_rank[nodename]
            for sname, rank in scenario_names_to_comm_rank.items():
                if sname in self.local_scenarios:
                    assert rank == comm.Get_rank()

        ## ensure we've set things up correctly for all local scenarios
        for sname in self.local_scenarios:
            for nodename, comm in self.comms.items():
                scenario_names_to_comm_rank = self.scenario_names_to_rank[nodename]
                if sname in scenario_names_to_comm_rank:
                    assert comm.Get_rank() == scenario_names_to_comm_rank[sname]


    def _compute_unconditional_node_probabilities(self):
        """ calculates unconditional node probabilities and prob_coeff
            and _PySP_W_coeff is set to a scalar 1 (used by variable_probability)"""
        for k,s in self.local_scenarios.items():
            root = s._mpisppy_node_list[0]
            root.uncond_prob = 1.0
            for parent,child in zip(s._mpisppy_node_list[:-1],s._mpisppy_node_list[1:]):
                child.uncond_prob = parent.uncond_prob * child.cond_prob
            if not hasattr(s._mpisppy_data, 'prob_coeff'):
                s._mpisppy_data.prob_coeff = dict()
                s._mpisppy_data.w_coeff = dict()
                for node in s._mpisppy_node_list:
                    s._mpisppy_data.prob_coeff[node.name] = (s._mpisppy_probability / node.uncond_prob)
                    s._mpisppy_data.w_coeff[node.name] = 1.0  # needs to be a float


    def _use_variable_probability_setter(self, verbose=False):
        """ set variable probability unconditional values using a function self.variable_probability
        that gives us a list of (id(vardata), probability)]
        ALSO set _PySP_W_coeff, which is a mask for W calculations (mask out zero probs)
        Note: We estimate that less than 0.01 of mpi-sppy runs will call this.
        """
        if self.variable_probability is None:
            for s in self.local_scenarios.values():
                s._mpisppy_data.has_variable_probability = False
            return
        didit = 0
        skipped = 0
        variable_probability_kwargs = self.options['variable_probability_kwargs'] \
                            if 'variable_probability_kwargs' in self.options \
                            else dict()
        for sname, s in self.local_scenarios.items():
            variable_probability = self.variable_probability(s, **variable_probability_kwargs)
            s._mpisppy_data.has_variable_probability = True
            for (vid, prob) in variable_probability:
                ndn, i = s._mpisppy_data.varid_to_nonant_index[vid]
                # If you are going to do any variables at a node, you have to do all.
                if type(s._mpisppy_data.prob_coeff[ndn]) is float:  # not yet a vector
                    defprob = s._mpisppy_data.prob_coeff[ndn]
                    s._mpisppy_data.prob_coeff[ndn] = np.full(s._mpisppy_data.nlens[ndn], defprob, dtype='d')
                    s._mpisppy_data.w_coeff[ndn] = np.ones(s._mpisppy_data.nlens[ndn], dtype='d')
                s._mpisppy_data.prob_coeff[ndn][i] = prob
                if prob == 0:  # there's probably a way to do this in numpy...
                    s._mpisppy_data.w_coeff[ndn][i] = 0
            didit += len(variable_probability)
            skipped += len(s._mpisppy_data.varid_to_nonant_index) - didit
        if verbose and self.cylinder_rank == 0:
            print ("variable_probability set",didit,"and skipped",skipped)

        if 'do_not_check_variable_probabilities' in self.options\
           and not self.options['do_not_check_variable_probabilities']:
            self._check_variable_probabilities_sum(verbose)

    def is_zero_prob( self, scenario_model, var ):
        """
        Args:
            scenario_model : a value in SPBase.local_scenarios
            var : a Pyomo Var on the scenario_model

        Returns:
            True if the variable has 0 probability, False otherwise
        """
        if self.variable_probability is None:
            return False
        _mpisppy_data = scenario_model._mpisppy_data
        ndn, i = _mpisppy_data.varid_to_nonant_index[id(var)]
        return float(_mpisppy_data.prob_coeff[ndn][i]) == 0.

    def _check_variable_probabilities_sum(self, verbose):

        nodenames = [] # to transmit to comms
        local_concats = {}   # keys are tree node names
        global_concats =  {} # values sums of node conditional probabilities

        # we need to accumulate all local contributions before the reduce
        for k,s in self.local_scenarios.items():
            nlens = s._mpisppy_data.nlens
            for node in s._mpisppy_node_list:
                if node.name not in nodenames:
                    ndn = node.name
                    nodenames.append(ndn)
                    local_concats[ndn] = np.zeros(nlens[ndn], dtype='d')
                    global_concats[ndn] = np.zeros(nlens[ndn], dtype='d')

        # sum local conditional probabilities
        for k,s in self.local_scenarios.items():
            for node in s._mpisppy_node_list:
                ndn = node.name
                local_concats[ndn] += s._mpisppy_data.prob_coeff[ndn]

        # compute sum node conditional probabilities (reduction)
        for ndn in nodenames:
            self.comms[ndn].Allreduce(
                [local_concats[ndn], MPI.DOUBLE],
                [global_concats[ndn], MPI.DOUBLE],
                op=MPI.SUM)

        tol = self.E1_tolerance
        checked_nodes = list()
        # check sum node conditional probabilites are close to 1
        for k,s in self.local_scenarios.items():
            nlens = s._mpisppy_data.nlens
            for node in s._mpisppy_node_list:
                ndn = node.name
                if ndn not in checked_nodes:
                    if not np.allclose(global_concats[ndn], 1., atol=tol):
                        notclose = ~np.isclose(global_concats[ndn], 1., atol=tol)
                        indices = np.nonzero(notclose)[0]
                        bad_vars = [ s._mpisppy_data.nonant_indices[ndn,idx].name for idx in indices ]
                        badprobs = [ global_concats[ndn][idx] for idx in indices]
                        raise RuntimeError(f"Node {ndn}, variables {bad_vars} have respective"
                                           f" conditional probability sum {badprobs}"
                                           " which are not 1")
                    checked_nodes.append(ndn)

    def _look_before_leap(self, scen, addlist):
        """ utility to check before attaching something to the user's model
        """
        for attr in addlist:
            if hasattr(scen, attr):
                raise RuntimeError("Model already has `internal' attribute" + attr)

    def _look_and_leap(self):
        for (sname, scenario) in self.local_scenarios.items():
            self._look_before_leap( scenario, ['_mpisppy_data', '_mpisppy_model' ] )

            scenario._mpisppy_data = pyo.Block(name="For non-Pyomo mpi-sppy data")
            scenario._mpisppy_model = pyo.Block(name="For mpi-sppy Pyomo additions to the scenario model")

            if hasattr(scenario, "PySP_prob"):
                raise RuntimeError(f"PySP_prob is deprecated; use _mpisppy_probability")
            if not hasattr(scenario, "_mpisppy_probability"):
                prob = 1./len(self.all_scenario_names)
                if self.cylinder_rank == 0:
                    print(f"Did not find _mpisppy_probability, assuming uniform probability {prob}")
                scenario._mpisppy_probability = prob
            if not hasattr(scenario, "_mpisppy_node_list"):
                raise RuntimeError(f"_mpisppy_node_list not found on scenario {sname}")

    def _options_check(self, required_options, given_options):
        """ Confirm that the specified list of options contains the specified
            list of required options. Raises a ValueError if anything is
            missing.
        """
        missing = [option for option in required_options if option not in given_options] 
        if missing:
            raise ValueError(f"Missing the following required options: {', '.join(missing)}")

    @property
    def spcomm(self):
        if self._spcomm is None:
            return None
        return self._spcomm()

    @spcomm.setter
    def spcomm(self, value):
        if self._spcomm is None:
            self._spcomm = weakref.ref(value)
        else:
            raise RuntimeError("SPBase.spcomm should only be set once")


    def gather_var_values_to_rank0(self, get_zero_prob_values=False):
        """ Gather the values of the nonanticipative variables to the root of
        the `mpicomm` for the cylinder

        Returns:
            dict or None:
                On the root (rank0), returns a dictionary mapping
                (scenario_name, variable_name) pairs to their values. On other
                ranks, returns None.
        """
        var_values = dict()
        for (sname, model) in self.local_scenarios.items():
            for node in model._mpisppy_node_list:
                for var in node.nonant_vardata_list:
                    var_name = var.name
                    if self.bundling:
                        dot_index = var_name.find('.')
                        assert dot_index >= 0
                        var_name = var_name[(dot_index+1):]
                    if (self.is_zero_prob(model, var)) and (not get_zero_prob_values):
                        var_values[sname, var_name] = None
                    else:
                        var_values[sname, var_name] = pyo.value(var)

        if self.n_proc == 1:
            return var_values

        result = self.mpicomm.gather(var_values, root=0)

        if (self.cylinder_rank == 0):
            result = {key: value
                for dic in result
                for (key, value) in dic.items()
            }
            return result


    def report_var_values_at_rank0(self, header="", print_zero_prob_values=False):
        """ Pretty-print the values and associated statistics for
        non-anticipative variables across all scenarios. """

        var_values = self.gather_var_values_to_rank0(get_zero_prob_values=print_zero_prob_values)

        if self.cylinder_rank == 0:

            if len(header) != 0:
                print(header)

            scenario_names = sorted(set(x for (x,y) in var_values))
            max_scenario_name_len = max(len(s) for s in scenario_names)
            variable_names = sorted(set(y for (x,y) in var_values))
            max_variable_name_len = max(len(v) for v in variable_names)
            # the "10" below is a reasonable minimum for floating-point output
            value_field_len = max(10, max_scenario_name_len)

            print("{0: <{width}s} | ".format("", width=max_variable_name_len), end='')
            for this_scenario in scenario_names:
                print("{0: ^{width}s} ".format(this_scenario, width=value_field_len), end='')
            print("")

            for this_var in variable_names:
                print("{0: <{width}} | ".format(this_var, width=max_variable_name_len), end='')
                for this_scenario in scenario_names:
                    this_var_value = var_values[this_scenario, this_var]
                    if (this_var_value == None) and (not print_zero_prob_values):
                        print("{0: ^{width}s}".format("-", width=value_field_len), end='')
                    else:
                        print("{0: {width}.4f}".format(this_var_value, width=value_field_len), end='')
                    print(" ", end='')
                print("")

    def write_first_stage_solution(self, file_name,
            first_stage_solution_writer=sputils.first_stage_nonant_writer):
        """ Writes the first-stage solution, if this object reports one available
            Args:
                file_name: path of file to write first stage solution to
                first_stage_solution_writer (optional): custom first stage solution writer function
        """

        if not self.first_stage_solution_available:
            raise RuntimeError("No first stage solution available")
        if self.cylinder_rank == 0:
            dirname = os.path.dirname(file_name)
            if dirname != '':
                os.makedirs(os.path.dirname(file_name), exist_ok=True)
            representative_scenario = self.local_scenarios[self.local_scenario_names[0]]
            first_stage_solution_writer(file_name, representative_scenario, self.bundling)

    def write_tree_solution(self, directory_name,
            scenario_tree_solution_writer=sputils.scenario_tree_solution_writer):
        """ Writes the tree solution, if this object reports one available.
            Raises a RuntimeError if it is not
            Args:
                directory_name: directory to write tree solution to
                scenario_tree_solution_writer (optional): custom scenario solution writer function
        """
        if not self.tree_solution_available:
            raise RuntimeError("No tree solution available")
        if self.cylinder_rank == 0:
            os.makedirs(directory_name, exist_ok=True)
        self.mpicomm.Barrier()
        for scenario_name, scenario in self.local_scenarios.items():
            scenario_tree_solution_writer(directory_name, scenario_name, scenario, self.bundling)



# Copyright 2020 by B. Knueven, D. Mildebrath, C. Muir, J-P Watson, and D.L. Woodruff
# This software is distributed under the 3-clause BSD License.

from pyomo.common.timing import TicTocTimer as _TTT

try:
    import mpi4py.MPI as _mpi
    haveMPI=True
except:
    haveMPI=False

tt_timer = _TTT()

if haveMPI:
    _global_rank = _mpi.COMM_WORLD.Get_rank()
else:
    _global_rank = 0

global_toc = lambda msg, cond=(_global_rank==0) : tt_timer.toc(msg, delta=False) if cond else None



from mpisppy.spbase import SPBase
import pyomo.environ as pyo
import logging
import parapint
from typing import List, Callable, Dict, Optional, Tuple, Any, Union
from pyomo.core.base.block import _BlockData
from pyomo.core.base.var import _GeneralVarData
from mpi4py import MPI
from mpisppy.utils.sputils import find_active_objective


logger = logging.getLogger('mpisppy.sc')


SCOptions = parapint.algorithms.IPOptions


def _assert_continuous(m: _BlockData):
    for v in m.component_data_objects(pyo.Var, descend_into=True, active=True):
        if not v.is_continuous():
            raise RuntimeError(f'Variable {v} in block {m} is not continuous; The Schur-Complement method only supports continuous problems.')


class _SCInterface(parapint.interfaces.MPIStochasticSchurComplementInteriorPointInterface):
    def __init__(self,
                 local_scenario_models: Dict[str, _BlockData],
                 all_scenario_names: List[str],
                 comm: MPI.Comm,
                 ownership_map: Dict):
        self.local_scenario_models = local_scenario_models

        models = list(local_scenario_models.values())
        ref_model = models[0]
        models = models[1:]
        self.nonant_vars = list(ref_model._mpisppy_data.nonant_indices.keys())

        super(_SCInterface, self).__init__(scenarios=all_scenario_names,
                                           nonanticipative_var_identifiers=self.nonant_vars,
                                           comm=comm,
                                           ownership_map=ownership_map)

    def build_model_for_scenario(self,
                                 scenario_identifier: str) -> Tuple[_BlockData, Dict[Any, _GeneralVarData]]:
        m = self.local_scenario_models[scenario_identifier]

        _assert_continuous(m)

        active_obj = find_active_objective(m)
        active_obj.deactivate()
        m._mpisppy_model.weighted_obj = pyo.Objective(expr=m._mpisppy_probability * active_obj.expr, sense=active_obj.sense)

        nonant_vars = m._mpisppy_data.nonant_indices
        if len(nonant_vars) != len(self.nonant_vars):
            raise ValueError(f'Number of non-anticipative variables is not consistent in scenario {scenario_identifier}.')

        return m, nonant_vars


class SchurComplement(SPBase):
    def __init__(self,
                 options: Union[Dict, SCOptions],
                 all_scenario_names: List,
                 scenario_creator: Callable,
                 scenario_creator_kwargs: Optional[Dict] = None,
                 all_nodenames=None,
                 mpicomm=None,
                 model_name=None,
                 suppress_warnings=False):
        super(SchurComplement, self).__init__(options=options,
                                              all_scenario_names=all_scenario_names,
                                              scenario_creator=scenario_creator,
                                              scenario_creator_kwargs=scenario_creator_kwargs,
                                              all_nodenames=all_nodenames,
                                              mpicomm=mpicomm)

        if self.bundling:
            raise ValueError('The Schur-Complement method does not support bundling')

        ownership_map = dict()
        for _rank, scenario_index_list in enumerate(self._rank_slices):
            for _scenario_ndx in scenario_index_list:
                ownership_map[_scenario_ndx] = _rank

        self.interface = _SCInterface(local_scenario_models=self.local_scenarios,
                                      all_scenario_names=self.all_scenario_names,
                                      comm=self.mpicomm,
                                      ownership_map=ownership_map)

    def solve(self):
        if isinstance(self.options, SCOptions):
            options = self.options()
        else:
            options = SCOptions()(self.options)
        if options.linalg.solver is None:
            options.linalg.solver = parapint.linalg.MPISchurComplementLinearSolver(
                subproblem_solvers={ndx: parapint.linalg.InteriorPointMA27Interface(cntl_options={1: 1e-6}) for ndx in range(len(self.all_scenario_names))},
                schur_complement_solver=parapint.linalg.InteriorPointMA27Interface(cntl_options={1: 1e-6}))

        status = parapint.algorithms.ip_solve(interface=self.interface,
                                              options=options)
        if status != parapint.algorithms.InteriorPointStatus.optimal:
            raise RuntimeError('Schur-Complement Interior Point algorithm did not converge')

        self.interface.load_primals_into_pyomo_model()

        return status



# Copyright 2020 by B. Knueven, D. Mildebrath, C. Muir, J-P Watson, and D.L. Woodruff
# This software is distributed under the 3-clause BSD License.
''' A template for creating PH_extension.py files
    NOTE: we pass in the ph object, so extensions can wreck everything
    if they want to!
    Written:  DLW Jan 2019
    Modified: DTM Aug 2019

    NOTE: The return values of all non-constructor methods are ignored
'''

class Extension:
    """ Abstract base class for extensions to general SPOpt objects.
    """
    def __init__(self, spopt_object):
        self.opt = spopt_object

    def pre_solve(self, subproblem):
        '''
        Method called before every subproblem solve

        Inputs
        ------
        subproblem : Pyomo subproblem (could be a scenario or bundle)

        Returns
        -------
        None
        '''
        pass

    def post_solve(self, subproblem, results):
        '''
        Method called after every subproblem solve

        Inputs
        ------
        subproblem : Pyomo subproblem (could be a scenario or bundle)
        results : Pyomo results object from initial solve or None if solve failed

        Returns
        -------
        results : Pyomo results objects from most recent solve
        '''
        return results

    def pre_iter0(self):
        ''' Method called at the end of PH_Prep().
            When this method is called, all scenarios have been created, and
            the dual/prox terms have been attached to the objective, but the
            solvers have not yet been created.
        '''
        pass

    def post_iter0(self):
        ''' Method called after the first PH iteration.
            When this method is called, one call to solve_loop() has been
            completed, and we have ensured that none of the models are
            infeasible. The rho_setter, if present, has not yet been applied.
        '''
        pass

    def miditer(self):
        ''' Method called after x-bar has been computed and the dual weights
            have been updated, but before solve_loop(). 
            If a converger is present, this method is called between the
            convergence_value() method and the is_converged() method.
        '''
        pass

    def enditer(self):
        ''' Method called after the solve_loop(), but before the next x-bar and
            weight update.
        '''
        pass

    def post_everything(self):
        ''' Method called after the termination of the algorithm.
            This method is called after the scenario_denouement, if a
            denouement is present. This function will not begin on any rank
            within self.opt.mpicomm until the scenario_denouement has completed
            on all other ranks.
        '''
        pass


class MultiExtension(Extension):
    """ Container for all the extension classes we are using.
        Also grabs ph and rank, so ad hoc calls (e.g., lagrangian) can use them.
    """
    def __init__(self, ph, ext_classes):
        super().__init__(ph)
        self.extdict = dict()

        # Construct multiple extension objects
        for constr in ext_classes:
            name = constr.__name__
            self.extdict[name] = constr(ph)

    def pre_iter0(self):
        for lobject in self.extdict.values():
            lobject.pre_iter0()
                                        
    def post_iter0(self):
        for lobject in self.extdict.values():
            lobject.post_iter0()

    def miditer(self):
        for lobject in self.extdict.values():
            lobject.miditer()

    def enditer(self):
        for lobject in self.extdict.values():
            lobject.enditer()

    def post_everything(self):
        for lobject in self.extdict.values():
            lobject.post_everything()



import unittest
import sys
import os
import parapint
from mpi4py import MPI


comm = MPI.COMM_WORLD
rank = comm.Get_rank()


class TestSC(unittest.TestCase):
    def setUp(self):
        self.original_path = sys.path
        example_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), '..', '..', 'examples', 'farmer')
        sys.path.append(example_dir)

    def tearDown(self):
        sys.path = self.original_path
    
    def test_farmer_example(self):
        import schur_complement as sc_example

        linear_solver = parapint.linalg.MPISchurComplementLinearSolver(subproblem_solvers={ndx: parapint.linalg.ScipyInterface(compute_inertia=True) for ndx in range(3)},
                                                                       schur_complement_solver=parapint.linalg.ScipyInterface(compute_inertia=True))
        sc_opt = sc_example.solve_with_sc(scen_count=3, linear_solver=linear_solver)
        sc_sol = sc_opt.gather_var_values_to_rank0()

        if rank == 0:
            self.assertAlmostEqual(sc_sol[('Scenario0', 'DevotedAcreage[CORN0]')], 80)
            self.assertAlmostEqual(sc_sol[('Scenario0', 'DevotedAcreage[SUGAR_BEETS0]')], 250)
            self.assertAlmostEqual(sc_sol[('Scenario0', 'DevotedAcreage[WHEAT0]')], 170)
            self.assertAlmostEqual(sc_sol[('Scenario1', 'DevotedAcreage[CORN0]')], 80)
            self.assertAlmostEqual(sc_sol[('Scenario1', 'DevotedAcreage[SUGAR_BEETS0]')], 250)
            self.assertAlmostEqual(sc_sol[('Scenario1', 'DevotedAcreage[WHEAT0]')], 170)
            self.assertAlmostEqual(sc_sol[('Scenario2', 'DevotedAcreage[CORN0]')], 80)
            self.assertAlmostEqual(sc_sol[('Scenario2', 'DevotedAcreage[SUGAR_BEETS0]')], 250)
            self.assertAlmostEqual(sc_sol[('Scenario2', 'DevotedAcreage[WHEAT0]')], 170)



# Put the license and copyright comments in py files as needed.
# If you are not sure you should run this, then you shouldn't.
# Not very robust; but probably good enough.
# Pipe to a file or run in an editor window so you can review the messages.
# NOTE: when it needs to update a file, it does so in-place.
# NOTE: the copyright line is not pep-8 compliant.

import os
import sys

if len(sys.argv) != 2:
    print ("Usage: python liccopy.py dirname")
    quit()

indir = sys.argv[1]
    
cstmt = "# Copyright 2020 by B. Knueven, D. Mildebrath, C. Muir, J-P Watson, and D.L. Woodruff"
lic = "# This software is distributed under the 3-clause BSD License."

pyomoscent = "#  Pyomo: Python Optimization Modeling Objects"


def _get_file(filepath):
    cline = -1
    licline = -1
    pyomofile = False
    with open(filepath,'r') as f:
        lines = f.readlines()
    for lno, line in enumerate(lines):
        if cstmt in line:
            cline = lno
            print
        if lic in line:
            licline = lno
        if pyomoscent in line:
            pyomofile = True
    return lines, cline, licline, pyomofile


def _write_file(filepath, lines, cline, licline, pyomofile):
    print (filepath, end='')

    if pyomofile:
        print(": seems to have Pyomo header; skipping")
        return
    
    cbefore = -1  # if only adding copyright
    licbefore = -1  # if only adding license
    if cline == -1 and licline != -1:
        cbefore = licline
    if cline != -1 and licline == -1:
        licbefore = licline + 1
    if cline != -1 and licline != -1:
        print(" .")
        return
    
    with open(filepath, "w") as f:
        if cline == -1 and licline == -1:
            f.write(cstmt+'\n')
            f.write(lic+'\n')
            print(": add both")
        for lno, line in enumerate(lines):
            if cbefore == lno:
                f.write(cstmt+'\n')
                print(": add copyright")
            if licbefore == lno:
                f.write(lic+'\n')
                print(": add license")
            f.write(line)


for subdir, dirs, files in os.walk(indir):
    for filename in files:
        filepath = subdir + os.sep + filename

        if filepath.endswith(".py"):
            lines, cline, licline, pyomofile = _get_file(filepath)
            _write_file(filepath, lines, cline, licline, pyomofile)
            



# Copyright 2020 by B. Knueven, D. Mildebrath, C. Muir, J-P Watson, and D.L. Woodruff
# This software is distributed under the 3-clause BSD License.
import mpisppy.spbase
import pyomo.environ as pyo
import logging
import mpisppy.utils.sputils as sputils

logger = logging.getLogger("mpisppy.ef")

class ExtensiveForm(mpisppy.spbase.SPBase):
    """ Create and solve an extensive form. 

    Attributes:
        ef (:class:`pyomo.environ.ConcreteModel`):
            Pyomo model of the extensive form.
        solver:
            Solver produced by the Pyomo solver factory.

    Args:
        options (dict):
            Dictionary of options. Must include a `solver` key to
            specify which solver to use on the EF.
        all_scenario_names (list):
            List of the names of each scenario in the EF (strings).
        scenario_creator (callable):
            Scenario creator function, which takes as input a scenario
            name, and returns a Pyomo model of that scenario.
        model_name (str, optional):
            Name of the resulting EF model object.
        scenario_creator_kwargs (dict):
            Keyword args passed to `scenario_creator`.
        suppress_warnings (bool, optional):
            Boolean to suppress warnings when building the EF. Default
            is False.
    """
    def __init__(
        self,
        options,
        all_scenario_names,
        scenario_creator,
        scenario_creator_kwargs=None,
        all_nodenames=None,
        model_name=None,
        suppress_warnings=False,
    ):
        """ Create the EF and associated solver. """
        super().__init__(
            options,
            all_scenario_names,
            scenario_creator,
            scenario_creator_kwargs=scenario_creator_kwargs,
            all_nodenames=all_nodenames
        )
        self.bundling = True
        if self.n_proc > 1 and self.cylinder_rank == 0:
            logger.warning("Creating an ExtensiveForm object in parallel. Why?")
        required = ["solver"]
        self._options_check(required, self.options)
        self.solver = pyo.SolverFactory(self.options["solver"])
        self.ef = sputils._create_EF_from_scen_dict(self.local_scenarios,
                EF_name=model_name)

    def solve_extensive_form(self, solver_options=None, tee=False):
        """ Solve the extensive form.
            
            Args:
                solver_options (dict, optional):
                    Dictionary of solver-specific options (e.g. Gurobi options,
                    CPLEX options, etc.).
                tee (bool, optional):
                    If True, displays solver output. Default False.

            Returns:
                :class:`pyomo.opt.results.results_.SolverResults`:
                    Result returned by the Pyomo solve method.
                
        """
        if "persistent" in self.options["solver"]:
            self.solver.set_instance(self.ef)
        # Pass solver-specifiec (e.g. Gurobi, CPLEX) options
        if solver_options is not None:
            for (opt, value) in solver_options.items():
                self.solver.options[opt] = value
        results = self.solver.solve(self.ef, tee=tee, load_solutions=False)
        if len(results.solution) > 0:
            if sputils.is_persistent(self.solver):
                self.solver.load_vars()
            else:
                self.ef.solutions.load_from(results)
            self.first_stage_solution_available = True
            self.tree_solution_available = True
        return results

    def get_objective_value(self):
        """ Retrieve the objective value.
        
        Returns:
            float:
                Objective value.

        Raises:
            ValueError:
                If optimal objective value could not be retrieved.
        """
        try:
            obj_val = pyo.value(self.ef.EF_Obj)
        except Exception as e:
            raise ValueError(f"Could not extract EF objective value with error: {str(e)}")
        return obj_val

    def get_root_solution(self):
        """ Get the value of the variables at the root node.

        Returns:
            dict:
                Dictionary mapping variable name (str) to variable value
                (float) for all variables at the root node.
        """
        result = dict()
        for var in self.ef.ref_vars.values():
            var_name = var.name
            dot_index = var_name.find(".")
            if dot_index >= 0 and var_name[:dot_index] in self.all_scenario_names:
                var_name = var_name[dot_index+1:]
            result[var_name] = var.value
        return result

    def nonants(self):
        """ An iterator to give representative Vars subject to non-anticipitivity
        Args: None

        Yields:
            tree node name, full EF Var name, Var value
        """
        yield from sputils.ef_nonants(self.ef)


    def nonants_to_csv(self, filename):
        """ Dump the nonant vars from an ef to a csv file; truly a dump...
        Args:
            filename (str): the full name of the csv output file
        """
        sputils.ef_nonants_csv(self.ef, filename)


    def scenarios(self):
        """ An iterator to give the scenario sub-models in an ef
        Args: None

        Yields:
            scenario name, scenario instance (str, ConcreteModel)
        """
        yield from self.local_scenarios.items()


if __name__ == "__main__":
    import mpisppy.tests.examples.farmer as farmer

    """ Farmer example """
    scenario_names = ["Scen" + str(i) for i in range(3)]
    scenario_creator_kwargs = {"sense": pyo.minimize, "use_integer": False}
    options = {"solver": "gurobi"}
    ef = ExtensiveForm(
        options,
        scenario_names,
        farmer.scenario_creator,
        model_name="TestEF",
        scenario_creator_kwargs=scenario_creator_kwargs,
    )
    results = ef.solve_extensive_form()
    print("Farmer objective value:", pyo.value(ef.ef.EF_Obj))



# Copyright 2020 by B. Knueven, D. Mildebrath, C. Muir, J-P Watson, and D.L. Woodruff
# This software is distributed under the 3-clause BSD License.
""" Plain versions of dictionaries that can be modified for each example
    as needed.
    ASSUME the corresponding args have been set up.
    IDIOM: we feel free to have unused dictionary entries."""

import copy

# Hub and spoke SPBase classes
from mpisppy.phbase import PHBase
from mpisppy.opt.ph import PH
from mpisppy.opt.aph import APH
from mpisppy.opt.lshaped import LShapedMethod
from mpisppy.fwph.fwph import FWPH
from mpisppy.utils.xhat_eval import Xhat_Eval
from mpisppy.cylinders.fwph_spoke import FrankWolfeOuterBound
from mpisppy.cylinders.lagrangian_bounder import LagrangianOuterBound
from mpisppy.cylinders.lagranger_bounder import LagrangerOuterBound
from mpisppy.cylinders.xhatlooper_bounder import XhatLooperInnerBound
from mpisppy.cylinders.xhatspecific_bounder import XhatSpecificInnerBound
from mpisppy.cylinders.xhatshufflelooper_bounder import XhatShuffleInnerBound
from mpisppy.cylinders.lshaped_bounder import XhatLShapedInnerBound
from mpisppy.cylinders.slam_heuristic import SlamUpHeuristic, SlamDownHeuristic
from mpisppy.cylinders.cross_scen_spoke import CrossScenarioCutSpoke
from mpisppy.cylinders.cross_scen_hub import CrossScenarioHub
from mpisppy.cylinders.hub import PHHub
from mpisppy.cylinders.hub import APHHub

def _hasit(args, argname):
    return hasattr(args, argname) and getattr(args, argname) is not None

def shared_options(args):
    shoptions = {
        "solvername": args.solver_name,
        "defaultPHrho": args.default_rho,
        "convthresh": 0,
        "PHIterLimit": args.max_iterations,  # not needed by all
        "verbose": args.with_verbose,
        "display_progress": args.with_display_progress,
        "display_convergence_detail": args.with_display_convergence_detail,
        "iter0_solver_options": dict(),
        "iterk_solver_options": dict(),
        "tee-rank0-solves": args.tee_rank0_solves,
        "trace_prefix" : args.trace_prefix,
    }
    if _hasit(args, "max_solver_threads"):
        shoptions["iter0_solver_options"]["threads"] = args.max_solver_threads
        shoptions["iterk_solver_options"]["threads"] = args.max_solver_threads
    if _hasit(args, "iter0_mipgap"):
        shoptions["iter0_solver_options"]["mipgap"] = args.iter0_mipgap
    if _hasit(args, "iterk_mipgap"):
        shoptions["iterk_solver_options"]["mipgap"] = args.iterk_mipgap
    return shoptions


def ph_hub(
    args,
    scenario_creator,
    scenario_denouement,
    all_scenario_names,
    scenario_creator_kwargs=None,
    ph_extensions=None,
    rho_setter=None,
    variable_probability=None,
):
    shoptions = shared_options(args)
    options = copy.deepcopy(shoptions)
    options["convthresh"] = args.intra_hub_conv_thresh
    options["bundles_per_rank"] = args.bundles_per_rank
    options["linearize_binary_proximal_terms"] = args.linearize_binary_proximal_terms
    options["linearize_proximal_terms"] = args.linearize_proximal_terms
    options["proximal_linearization_tolerance"] = args.proximal_linearization_tolerance

    if _hasit(args, "with_cross_scenario_cuts") and args.with_cross_scenario_cuts:
        hub_class = CrossScenarioHub
    else:
        hub_class = PHHub

    hub_dict = {
        "hub_class": hub_class,
        "hub_kwargs": {"options": {"rel_gap": args.rel_gap,
                                   "abs_gap": args.abs_gap}},
        "opt_class": PH,
        "opt_kwargs": {
            "options": options,
            "all_scenario_names": all_scenario_names,
            "scenario_creator": scenario_creator,
            "scenario_creator_kwargs": scenario_creator_kwargs,
            "scenario_denouement": scenario_denouement,
            "rho_setter": rho_setter,
            "variable_probability": variable_probability,
            "extensions": ph_extensions,
        }
    }
    return hub_dict


def aph_hub(
    args,
    scenario_creator,
    scenario_denouement,
    all_scenario_names,
    scenario_creator_kwargs=None,
    ph_extensions=None,
    rho_setter=None,
    variable_probability=None,
):
    hub_dict = ph_hub(args,
                      scenario_creator,
                      scenario_denouement,
                      all_scenario_names,
                      scenario_creator_kwargs=scenario_creator_kwargs,
                      ph_extensions=ph_extensions,
                      rho_setter=rho_setter,
                      variable_probability=variable_probability)

    hub_dict['hub_class'] = APHHub
    hub_dict['opt_class'] = APH    

    hub_dict['opt_kwargs']['options']['APHgamma'] = args.aph_gamma
    hub_dict['opt_kwargs']['options']['APHnu'] = args.aph_nu
    hub_dict['opt_kwargs']['options']['async_frac_needed'] = args.aph_frac_needed
    hub_dict['opt_kwargs']['options']['dispatch_frac'] = args.aph_dispatch_frac
    hub_dict['opt_kwargs']['options']['async_sleep_secs'] = args.aph_sleep_seconds

    return hub_dict


def fwph_spoke(
    args,
    scenario_creator,
    scenario_denouement,
    all_scenario_names,
    scenario_creator_kwargs=None,
):
    shoptions = shared_options(args)

    mip_solver_options, qp_solver_options = dict(), dict()
    if _hasit(args, "max_solver_threads"):
        mip_solver_options["threads"] = args.max_solver_threads
        qp_solver_options["threads"] = args.max_solver_threads
    if _hasit(args, "fwph_mipgap"):
        mip_solver_options["mipgap"] = args.fwph_mipgap

    fw_options = {
        "FW_iter_limit": args.fwph_iter_limit,
        "FW_weight": args.fwph_weight,
        "FW_conv_thresh": args.fwph_conv_thresh,
        "stop_check_tol": args.fwph_stop_check_tol,
        "solvername": args.solver_name,
        "FW_verbose": args.with_verbose,
        "mip_solver_options" : mip_solver_options,
        "qp_solver_options" : qp_solver_options,
    }
    fw_dict = {
        "spoke_class": FrankWolfeOuterBound,
        "spoke_kwargs": dict(),
        "opt_class": FWPH,
        "opt_kwargs": {
            "PH_options": shoptions,  # be sure convthresh is zero for fwph
            "FW_options": fw_options,
            "all_scenario_names": all_scenario_names,
            "scenario_creator": scenario_creator,
            "scenario_creator_kwargs": scenario_creator_kwargs,
            "scenario_denouement": scenario_denouement            
        },
    }
    return fw_dict


def lagrangian_spoke(
    args,
    scenario_creator,
    scenario_denouement,
    all_scenario_names,
    scenario_creator_kwargs=None,
    rho_setter=None,
):
    shoptions = shared_options(args)
    lagrangian_spoke = {
        "spoke_class": LagrangianOuterBound,
        "spoke_kwargs": dict(),
        "opt_class": PHBase,
        "opt_kwargs": {
            "options": shoptions,
            "all_scenario_names": all_scenario_names,
            "scenario_creator": scenario_creator,
            "scenario_creator_kwargs": scenario_creator_kwargs,
            'scenario_denouement': scenario_denouement,            
            "rho_setter": rho_setter

        }
    }
    if args.lagrangian_iter0_mipgap is not None:
        lagrangian_spoke["opt_kwargs"]["options"]["iter0_solver_options"]\
            ["mipgap"] = args.lagrangian_iter0_mipgap
    if args.lagrangian_iterk_mipgap is not None:
        lagrangian_spoke["opt_kwargs"]["options"]["iterk_solver_options"]\
            ["mipgap"] = args.lagrangian_iterk_mipgap
    return lagrangian_spoke


# special lagrangian that computes its own xhat and W
def lagranger_spoke(
    args,
    scenario_creator,
    scenario_denouement,
    all_scenario_names,
    scenario_creator_kwargs=None,
    rho_setter=None
):
    shoptions = shared_options(args)
    lagranger_spoke = {
        "spoke_class": LagrangerOuterBound,
        "spoke_kwargs": dict(),
        "opt_class": PHBase,
        "opt_kwargs": {
            "options": shoptions,
            "all_scenario_names": all_scenario_names,
            "scenario_creator": scenario_creator,
            "scenario_creator_kwargs": scenario_creator_kwargs,
            'scenario_denouement': scenario_denouement,            
            "rho_setter": rho_setter
        }
    }
    if args.lagranger_iter0_mipgap is not None:
        lagranger_spoke["opt_kwargs"]["options"]["iter0_solver_options"]\
            ["mipgap"] = args.lagranger_iter0_mipgap
    if args.lagranger_iterk_mipgap is not None:
        lagranger_spoke["opt_kwargs"]["options"]["iterk_solver_options"]\
            ["mipgap"] = args.lagranger_iterk_mipgap
    if args.lagranger_rho_rescale_factors_json is not None:
        lagranger_spoke["opt_kwargs"]["options"]\
            ["lagranger_rho_rescale_factors_json"]\
            = args.lagranger_rho_rescale_factors_json
        
    return lagranger_spoke

        
def xhatlooper_spoke(
    args,
    scenario_creator,
    scenario_denouement,
    all_scenario_names,
    scenario_creator_kwargs=None,
):
    
    shoptions = shared_options(args)
    xhat_options = copy.deepcopy(shoptions)
    xhat_options['bundles_per_rank'] = 0 #  no bundles for xhat
    xhat_options["xhat_looper_options"] = {
        "xhat_solver_options": shoptions["iterk_solver_options"],
        "scen_limit": args.xhat_scen_limit,
        "dump_prefix": "delme",
        "csvname": "looper.csv",
    }
    xhatlooper_dict = {
        "spoke_class": XhatLooperInnerBound,
        "spoke_kwargs": dict(),
        "opt_class": Xhat_Eval,
        "opt_kwargs": {
            "options": xhat_options,
            "all_scenario_names": all_scenario_names,
            "scenario_creator": scenario_creator,
            "scenario_creator_kwargs": scenario_creator_kwargs,
            "scenario_denouement": scenario_denouement            
        },
    }
    return xhatlooper_dict


def xhatshuffle_spoke(
    args,
    scenario_creator,
    scenario_denouement,
    all_scenario_names,
    scenario_creator_kwargs=None,
):

    shoptions = shared_options(args)
    xhat_options = copy.deepcopy(shoptions)
    xhat_options['bundles_per_rank'] = 0 #  no bundles for xhat
    xhat_options["xhat_looper_options"] = {
        "xhat_solver_options": shoptions["iterk_solver_options"],
        "dump_prefix": "delme",
        "csvname": "looper.csv",
    }
    xhatlooper_dict = {
        "spoke_class": XhatShuffleInnerBound,
        "spoke_kwargs": dict(),
        "opt_class": Xhat_Eval,
        "opt_kwargs": {
            "options": xhat_options,
            "all_scenario_names": all_scenario_names,
            "scenario_creator": scenario_creator,
            "scenario_creator_kwargs": scenario_creator_kwargs,
            "scenario_denouement": scenario_denouement                        
        },
    }
    return xhatlooper_dict


def xhatspecific_spoke(
    args,
    scenario_creator,
    scenario_denouement,
    all_scenario_names,
    scenario_dict,
    all_nodenames=None,
    BFs=None,
    scenario_creator_kwargs=None,
):
    
    shoptions = shared_options(args)
    xhat_options = copy.deepcopy(shoptions)
    xhat_options["xhat_specific_options"] = {
        "xhat_solver_options": shoptions["iterk_solver_options"],
        "xhat_scenario_dict": scenario_dict,
        "csvname": "specific.csv",
    }
    if BFs:
        xhat_options["branching_factors"] = BFs

    xhat_options['bundles_per_rank'] = 0 #  no bundles for xhat
    xhatspecific_dict = {
        "spoke_class": XhatSpecificInnerBound,
        "spoke_kwargs": dict(),
        "opt_class": Xhat_Eval,
        "opt_kwargs": {
            "options": xhat_options,
            "all_scenario_names": all_scenario_names,
            "scenario_creator": scenario_creator,
            "scenario_creator_kwargs": scenario_creator_kwargs,
            "scenario_denouement": scenario_denouement,
            "all_nodenames": all_nodenames,
        },
    }
    return xhatspecific_dict

def xhatlshaped_spoke(
    args,
    scenario_creator,
    scenario_denouement,
    all_scenario_names,
    scenario_creator_kwargs=None,
):
    
    shoptions = shared_options(args)
    xhat_options = copy.deepcopy(shoptions)
    xhat_options['bundles_per_rank'] = 0 #  no bundles for xhat

    xhatlshaped_dict = {
        "spoke_class": XhatLShapedInnerBound,
        "spoke_kwargs": dict(),
        "opt_class": Xhat_Eval,
        "opt_kwargs": {
            "options": xhat_options,
            "all_scenario_names": all_scenario_names,
            "scenario_creator": scenario_creator,
            "scenario_creator_kwargs": scenario_creator_kwargs,
            "scenario_denouement": scenario_denouement            
        },
    }
    return xhatlshaped_dict

def slamup_spoke(
    args,
    scenario_creator,
    scenario_denouement,
    all_scenario_names,
    scenario_creator_kwargs=None,
):

    shoptions = shared_options(args)
    xhat_options = copy.deepcopy(shoptions)
    xhat_options['bundles_per_rank'] = 0 #  no bundles for xhat
    xhatlooper_dict = {
        "spoke_class": SlamUpHeuristic,
        "spoke_kwargs": dict(),
        "opt_class": Xhat_Eval,
        "opt_kwargs": {
            "options": xhat_options,
            "all_scenario_names": all_scenario_names,
            "scenario_creator": scenario_creator,
            "scenario_creator_kwargs": scenario_creator_kwargs,
            "scenario_denouement": scenario_denouement
        },
    }
    return xhatlooper_dict

def slamdown_spoke(
    args,
    scenario_creator,
    scenario_denouement,
    all_scenario_names,
    scenario_creator_kwargs=None,
):

    shoptions = shared_options(args)
    xhat_options = copy.deepcopy(shoptions)
    xhat_options['bundles_per_rank'] = 0 #  no bundles for xhat
    xhatlooper_dict = {
        "spoke_class": SlamDownHeuristic,
        "spoke_kwargs": dict(),
        "opt_class": Xhat_Eval,
        "opt_kwargs": {
            "options": xhat_options,
            "all_scenario_names": all_scenario_names,
            "scenario_creator": scenario_creator,
            "scenario_creator_kwargs": scenario_creator_kwargs,
            "scenario_denouement": scenario_denouement            
        },
    }
    return xhatlooper_dict

def cross_scenario_cut_spoke(
    args,
    scenario_creator,
    scenario_denouement,
    all_scenario_names,
    scenario_creator_kwargs=None,
):

    if _hasit(args, "max_solver_threads"):
        sp_solver_options = {"threads":args.max_solver_threads}
    else:
        sp_solver_options = dict() 

    if _hasit(args, "eta_bounds_mipgap"):
        sp_solver_options["mipgap"] = args.eta_bounds_mipgap

    ls_options = { "root_solver" : args.solver_name,
                   "sp_solver": args.solver_name,
                   "sp_solver_options" : sp_solver_options,
                    "verbose": args.with_verbose,
                 }
    cut_spoke = {
        "spoke_class": CrossScenarioCutSpoke,
        "spoke_kwargs": dict(),
        "opt_class": LShapedMethod,
        "opt_kwargs": {
            "options": ls_options,
            "all_scenario_names": all_scenario_names,
            "scenario_creator": scenario_creator,
            "scenario_creator_kwargs": scenario_creator_kwargs,
            "scenario_denouement": scenario_denouement            
            },
        }

    return cut_spoke



# Copyright 2020 by B. Knueven, D. Mildebrath, C. Muir, J-P Watson, and D.L. Woodruff
# This software is distributed under the 3-clause BSD License.
''' An extension to save PH weights and/or PH xbar values to csv files.

    To use, specify either or both of the following keys to the options dict:

        "W_fname" : <str filename>
        "Xbar_fname": <str filename>

    If neither option is specified, this extension does nothing (prints a
    warning). If only one is specified, then only those values are saved.
    The resulting files will be output in the format:

        (W values) csv with rows: scenario_name, variable_name, weight_value
        (x values) csv with rows: variable_name, variable_value

    This format is consistent with the input required by WXbarReader.

    TODO:
        Check with bundles

    Written: DLW, July 2019
    Modified: DTM, Aug 2019
'''

import os
import pyomo.environ as pyo
import mpisppy.utils.wxbarutils
import mpisppy.extensions.extension

class WXBarWriter(mpisppy.extensions.extension.Extension):
    """ Extension class for writing the W values
    """
    def __init__(self, ph, rank, n_proc):
        # Check a bunch of files
        w_fname, x_fname, sep_files = None, None, False

        if ('W_fname' in ph.options): # W_fname is a path if separate_W_files=True
            w_fname = ph.options['W_fname']
        if ('Xbar_fname' in ph.options):
            x_fname = ph.options['Xbar_fname']
        if ('separate_W_files' in ph.options):
            sep_files = ph.options['separate_W_files']

        if (x_fname is None and w_fname is None and rank==0):
            print('Warning: no output files provided to WXBarWriter. '
                  'No values will be saved.')

        if (w_fname and (not sep_files) and os.path.exists(w_fname) and rank==0):
            print('Warning: specified W_fname ({fn})'.format(fn=w_fname) +
                  ' already exists. Results will be appended to this file.')
        elif (w_fname and sep_files and (not os.path.exists(w_fname)) and rank==0):
            print('Warning: path {path} does not exist. Creating...'.format(
                    path=w_fname))
            os.makedirs(w_fname, exist_ok=True)

        if (x_fname and os.path.exists(x_fname) and rank==0):
            print('Warning: specified Xbar_fname ({fn})'.format(fn=x_fname) +
                  ' already exists. Results will be appended to this file.')

        self.PHB = ph
        self.cylinder_rank = rank
        self.w_fname = w_fname
        self.x_fname = x_fname
        self.sep_files = sep_files # Write separate files for each 
                                   # scenario's dual weights

    def pre_iter0(self):
        pass

    def post_iter0(self):
        pass
        
    def miditer(self, PHIter, conv):
        pass

    def enditer(self, PHIter):
        # Maybe you want to write here, but with an iteration specific name,
        # and maybe just for some iterations.
        pass

    def post_everything(self, PHIter, conv):
        if (self.w_fname):
            mpisppy.utils.wxbarutils.write_W_to_file(self.PHB, self.w_fname,
                sep_files=self.sep_files)
        if (self.x_fname):
            mpisppy.utils.wxbarutils.write_xbar_to_file(self.PHB, self.x_fname)




# Copyright 2020 by B. Knueven, D. Mildebrath, C. Muir, J-P Watson, and D.L. Woodruff
# This software is distributed under the 3-clause BSD License.
import mpisppy.cylinders.spoke

class FrankWolfeOuterBound(mpisppy.cylinders.spoke.OuterBoundSpoke):

    converger_spoke_char = 'F'

    def main(self):
        self.opt.fwph_main()

    def is_converged(self):
        return self.got_kill_signal()

    def sync(self):
        # Tell the hub about the most recent bound
        self.bound = self.opt._local_bound

    def finalize(self):
        # The FWPH spoke calls "is_converged" before it
        # even starts doing anything, so its possible
        # to get here without any bound information
        # if we terminated early
        if not hasattr(self.opt, 'local_bound'):
            return
        self.bound = self.opt._local_bound
        self.final_bound = self.opt._local_bound
        return self.final_bound
